# Copyright (c) 2012-2013 Mitch Garnaat http://garnaat.org/
# Copyright 2012-2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import logging
import os
import threading
import time

from botocore import parsers
from botocore.awsrequest import create_request_object
from botocore.exceptions import HTTPClientError
from botocore.history import get_global_history_recorder
from botocore.hooks import first_non_none_response
from botocore.httpchecksum import handle_checksum_body
from botocore.httpsession import URLLib3Session
from botocore.response import StreamingBody
from botocore.utils import (
    get_environ_proxies,
    is_valid_endpoint_url,
    is_valid_ipv6_endpoint_url,
)

logger = logging.getLogger(__name__)
history_recorder = get_global_history_recorder()
DEFAULT_TIMEOUT = 60
MAX_POOL_CONNECTIONS = 10


def convert_to_response_dict(http_response, operation_model):
    """Convert an HTTP response object to a request dict.

    This converts the requests library's HTTP response object to
    a dictionary.

    :type http_response: botocore.vendored.requests.model.Response
    :param http_response: The HTTP response from an AWS service request.

    :rtype: dict
    :return: A response dictionary which will contain the following keys:
        * headers (dict)
        * status_code (int)
        * body (string or file-like object)

    """
    response_dict = {
        'headers': http_response.headers,
        'status_code': http_response.status_code,
        'context': {
            'operation_name': operation_model.name,
        },
    }
    if response_dict['status_code'] >= 300:
        response_dict['body'] = http_response.content
    elif operation_model.has_event_stream_output:
        response_dict['body'] = http_response.raw
    elif operation_model.has_streaming_output:
        length = response_dict['headers'].get('content-length')
        response_dict['body'] = StreamingBody(http_response.raw, length)
    else:
        response_dict['body'] = http_response.content
    return response_dict


class Endpoint:
    """
    Represents an endpoint for a particular service in a specific
    region.  Only an endpoint can make requests.

    :ivar service: The Service object that describes this endpoints
        service.
    :ivar host: The fully qualified endpoint hostname.
    :ivar session: The session object.
    """

    def __init__(
        self,
        host,
        endpoint_prefix,
        event_emitter,
        response_parser_factory=None,
        http_session=None,
    ):
        self._endpoint_prefix = endpoint_prefix
        self._event_emitter = event_emitter
        self.host = host
        self._lock = threading.Lock()
        if response_parser_factory is None:
            response_parser_factory = parsers.ResponseParserFactory()
        self._response_parser_factory = response_parser_factory
        self.http_session = http_session
        if self.http_session is None:
            self.http_session = URLLib3Session()

    def __repr__(self):
        return f'{self._endpoint_prefix}({self.host})'

    def make_request(self, operation_model, request_dict):
        logger.debug(
            "Making request for %s with params: %s",
            operation_model,
            request_dict,
        )
        return self._send_request(request_dict, operation_model)

    def create_request(self, params, operation_model=None):
        request = create_request_object(params)
        if operation_model:
            request.stream_output = any(
                [
                    operation_model.has_streaming_output,
                    operation_model.has_event_stream_output,
                ]
            )
            service_id = operation_model.service_model.service_id.hyphenize()
            event_name = f'request-created.{service_id}.{operation_model.name}'
            self._event_emitter.emit(
                event_name,
                request=request,
                operation_name=operation_model.name,
            )
        prepared_request = self.prepare_request(request)
        return prepared_request

    def _encode_headers(self, headers):
        # In place encoding of headers to utf-8 if they are unicode.
        for key, value in headers.items():
            if isinstance(value, str):
                headers[key] = value.encode('utf-8')

    def prepare_request(self, request):
        self._encode_headers(request.headers)
        return request.prepare()

    def _send_request(self, request_dict, operation_model):
        attempts = 1
        request = self.create_request(request_dict, operation_model)
        context = request_dict['context']
        success_response, exception = self._get_response(
            request, operation_model, context
        )
        while self._needs_retry(
            attempts,
            operation_model,
            request_dict,
            success_response,
            exception,
        ):
            attempts += 1
            # If there is a stream associated with the request, we need
            # to reset it before attempting to send the request again.
            # This will ensure that we resend the entire contents of the
            # body.
            request.reset_stream()
            # Create a new request when retried (including a new signature).
            request = self.create_request(request_dict, operation_model)
            success_response, exception = self._get_response(
                request, operation_model, context
            )
        if (
            success_response is not None
            and 'ResponseMetadata' in success_response[1]
        ):
            # We want to share num retries, not num attempts.
            total_retries = attempts - 1
            success_response[1]['ResponseMetadata']['RetryAttempts'] = (
                total_retries
            )
        if exception is not None:
            raise exception
        else:
            return success_response

    def _get_response(self, request, operation_model, context):
        # This will return a tuple of (success_response, exception)
        # and success_response is itself a tuple of
        # (http_response, parsed_dict).
        # If an exception occurs then the success_response is None.
        # If no exception occurs then exception is None.
        success_response, exception = self._do_get_response(
            request, operation_model, context
        )
        kwargs_to_emit = {
            'response_dict': None,
            'parsed_response': None,
            'context': context,
            'exception': exception,
        }
        if success_response is not None:
            http_response, parsed_response = success_response
            kwargs_to_emit['parsed_response'] = parsed_response
            kwargs_to_emit['response_dict'] = convert_to_response_dict(
                http_response, operation_model
            )
        service_id = operation_model.service_model.service_id.hyphenize()
        self._event_emitter.emit(
            f'response-received.{service_id}.{operation_model.name}',
            **kwargs_to_emit,
        )
        return success_response, exception

    def _do_get_response(self, request, operation_model, context):
        try:
            logger.debug("Sending http request: %s", request)
            history_recorder.record(
                'HTTP_REQUEST',
                {
                    'method': request.method,
                    'headers': request.headers,
                    'streaming': operation_model.has_streaming_input,
                    'url': request.url,
                    'body': request.body,
                },
            )
            service_id = operation_model.service_model.service_id.hyphenize()
            event_name = f'before-send.{service_id}.{operation_model.name}'
            responses = self._event_emitter.emit(event_name, request=request)
            http_response = first_non_none_response(responses)
            if http_response is None:
                http_response = self._send(request)
        except HTTPClientError as e:
            return (None, e)
        except Exception as e:
            logger.debug(
                "Exception received when sending HTTP request.", exc_info=True
            )
            return (None, e)
        # This returns the http_response and the parsed_data.
        response_dict = convert_to_response_dict(
            http_response, operation_model
        )
        handle_checksum_body(
            http_response,
            response_dict,
            context,
            operation_model,
        )

        http_response_record_dict = response_dict.copy()
        http_response_record_dict['streaming'] = (
            operation_model.has_streaming_output
        )
        history_recorder.record('HTTP_RESPONSE', http_response_record_dict)

        protocol = operation_model.service_model.resolved_protocol
        customized_response_dict = {}
        self._event_emitter.emit(
            f"before-parse.{service_id}.{operation_model.name}",
            operation_model=operation_model,
            response_dict=response_dict,
            customized_response_dict=customized_response_dict,
        )
        parser = self._response_parser_factory.create_parser(protocol)
        parsed_response = parser.parse(
            response_dict, operation_model.output_shape
        )
        parsed_response.update(customized_response_dict)
        # Do a second parsing pass to pick up on any modeled error fields
        # NOTE: Ideally, we would push this down into the parser classes but
        # they currently have no reference to the operation or service model
        # The parsers should probably take the operation model instead of
        # output shape but we can't change that now
        if http_response.status_code >= 300:
            self._add_modeled_error_fields(
                response_dict,
                parsed_response,
                operation_model,
                parser,
            )
        history_recorder.record('PARSED_RESPONSE', parsed_response)
        return (http_response, parsed_response), None

    def _add_modeled_error_fields(
        self,
        response_dict,
        parsed_response,
        operation_model,
        parser,
    ):
        error_code = parsed_response.get("Error", {}).get("Code")
        if error_code is None:
            return
        service_model = operation_model.service_model
        error_shape = service_model.shape_for_error_code(error_code)
        if error_shape is None:
            return
        modeled_parse = parser.parse(response_dict, error_shape)
        # TODO: avoid naming conflicts with ResponseMetadata and Error
        parsed_response.update(modeled_parse)

    def _needs_retry(
        self,
        attempts,
        operation_model,
        request_dict,
        response=None,
        caught_exception=None,
    ):
        service_id = operation_model.service_model.service_id.hyphenize()
        event_name = f'needs-retry.{service_id}.{operation_model.name}'
        responses = self._event_emitter.emit(
            event_name,
            response=response,
            endpoint=self,
            operation=operation_model,
            attempts=attempts,
            caught_exception=caught_exception,
            request_dict=request_dict,
        )
        handler_response = first_non_none_response(responses)
        if handler_response is None:
            return False
        else:
            # Request needs to be retried, and we need to sleep
            # for the specified number of times.
            logger.debug(
                "Response received to retry, sleeping for " "%s seconds",
                handler_response,
            )
            time.sleep(handler_response)
            return True

    def _send(self, request):
        return self.http_session.send(request)


class EndpointCreator:
    def __init__(self, event_emitter):
        self._event_emitter = event_emitter

    def create_endpoint(
        self,
        service_model,
        region_name,
        endpoint_url,
        verify=None,
        response_parser_factory=None,
        timeout=DEFAULT_TIMEOUT,
        max_pool_connections=MAX_POOL_CONNECTIONS,
        http_session_cls=URLLib3Session,
        proxies=None,
        socket_options=None,
        client_cert=None,
        proxies_config=None,
    ):
        if not is_valid_endpoint_url(
            endpoint_url
        ) and not is_valid_ipv6_endpoint_url(endpoint_url):
            raise ValueError(f"Invalid endpoint: {endpoint_url}")

        if proxies is None:
            proxies = self._get_proxies(endpoint_url)
        endpoint_prefix = service_model.endpoint_prefix

        logger.debug('Setting %s timeout as %s', endpoint_prefix, timeout)
        http_session = http_session_cls(
            timeout=timeout,
            proxies=proxies,
            verify=self._get_verify_value(verify),
            max_pool_connections=max_pool_connections,
            socket_options=socket_options,
            client_cert=client_cert,
            proxies_config=proxies_config,
        )

        return Endpoint(
            endpoint_url,
            endpoint_prefix=endpoint_prefix,
            event_emitter=self._event_emitter,
            response_parser_factory=response_parser_factory,
            http_session=http_session,
        )

    def _get_proxies(self, url):
        # We could also support getting proxies from a config file,
        # but for now proxy support is taken from the environment.
        return get_environ_proxies(url)

    def _get_verify_value(self, verify):
        # This is to account for:
        # https://github.com/kennethreitz/requests/issues/1436
        # where we need to honor REQUESTS_CA_BUNDLE because we're creating our
        # own request objects.
        # First, if verify is not None, then the user explicitly specified
        # a value so this automatically wins.
        if verify is not None:
            return verify
        # Otherwise use the value from REQUESTS_CA_BUNDLE, or default to
        # True if the env var does not exist.
        return os.environ.get('REQUESTS_CA_BUNDLE', True)



# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import json

from ruamel.yaml import YAML
from ruamel.yaml.error import YAMLError

from awscli.argprocess import ParamError, ParamSyntaxError
from awscli.customizations.arguments import OverrideRequiredArgsArgument
from awscli.paramfile import LOCAL_PREFIX_MAP, get_paramfile


def register_cli_input_args(cli):
    cli.register('building-argument-table', add_cli_input_json)
    cli.register('building-argument-table', add_cli_input_yaml)


def add_cli_input_json(session, argument_table, **kwargs):
    _add_cli_input_argument(session, argument_table, CliInputJSONArgument)


def add_cli_input_yaml(session, argument_table, **kwargs):
    _add_cli_input_argument(session, argument_table, CliInputYAMLArgument)


def _add_cli_input_argument(session, argument_table, argument_cls):
    # This argument cannot support operations with streaming output which
    # is designated by the argument name `outfile`.
    if 'outfile' not in argument_table:
        cli_input_argument = argument_cls(session)
        cli_input_argument.add_to_arg_table(argument_table)


class CliInputArgument(OverrideRequiredArgsArgument):
    """This argument inputs a formatted string as the entire command input.

    Ideally, the value to this argument will be a filled out file in a
    format such as JSON that was generated by ``--generate-cli-skeleton``.
    The parameters in the file will be overwritten by any arguments specified
    on the command line.
    """

    def _register_argument_action(self):
        self._session.register(
            'calling-command.*', self.add_to_call_parameters
        )
        super(CliInputArgument, self)._register_argument_action()

    def add_to_call_parameters(self, call_parameters, parsed_args, **kwargs):
        arg_value = self._get_arg_value(parsed_args)
        if arg_value is None:
            return

        loaded_params = self._load_parameters(arg_value)
        if not isinstance(loaded_params, dict):
            raise ParamError(
                self.cli_name,
                "Invalid type: expecting map, "
                "received %s" % type(loaded_params),
            )
        self._update_call_parameters(call_parameters, loaded_params)

    def _get_arg_value(self, parsed_args):
        arg_value = getattr(parsed_args, self.py_name, None)
        if arg_value is None:
            return

        cli_input_args = [
            k
            for k, v in vars(parsed_args).items()
            if v is not None and k.startswith('cli_input')
        ]
        if len(cli_input_args) != 1:
            raise ParamSyntaxError(
                'Only one --cli-input- parameter may be specified.'
            )

        # If the value starts with file:// or fileb://, return the contents
        # from the file.
        paramfile_data = get_paramfile(arg_value, LOCAL_PREFIX_MAP)
        if paramfile_data is not None:
            return paramfile_data

        return arg_value

    def _load_parameters(self, arg_value):
        raise NotImplementedError('_load_parameters')

    def _update_call_parameters(self, call_parameters, loaded_parameters):
        for input_key in loaded_parameters.keys():
            # Only add the values to ``call_parameters`` if not already
            # present.
            if input_key not in call_parameters:
                call_parameters[input_key] = loaded_parameters[input_key]


class CliInputJSONArgument(CliInputArgument):
    """This argument inputs a JSON string as the entire input for a command.

    Ideally, the value to this argument should be a filled out JSON file
    generated by ``--generate-cli-skeleton``. The items in the JSON string
    will not clobber other arguments entered into the command line.
    """

    ARG_DATA = {
        'name': 'cli-input-json',
        'group_name': 'cli_input',
        'help_text': (
            'Reads arguments from the JSON string provided. The JSON string '
            'follows the format provided by ``--generate-cli-skeleton``. If '
            'other arguments are provided on the command line, those values '
            'will override the JSON-provided values. It is not possible to '
            'pass arbitrary binary values using a JSON-provided value as the '
            'string will be taken literally. This may not be specified along '
            'with ``--cli-input-yaml``.'
        ),
    }

    def _load_parameters(self, arg_value):
        try:
            return json.loads(arg_value)
        except ValueError:
            raise ParamError(self.name, "Invalid JSON received.")


class CliInputYAMLArgument(CliInputArgument):
    ARG_DATA = {
        'name': 'cli-input-yaml',
        'group_name': 'cli_input',
        'help_text': (
            'Reads arguments from the YAML string provided. The YAML string '
            'follows the format provided by '
            '``--generate-cli-skeleton yaml-input``. '
            'If other arguments are provided on the command line, those '
            'values will override the YAML-provided values. This may not be '
            'specified along with ``--cli-input-json``.'
        ),
    }

    def _load_parameters(self, arg_value):
        yaml = YAML(typ='safe', pure=True)

        try:
            return yaml.load(arg_value)
        except YAMLError:
            raise ParamError(self.name, "Invalid YAML received.")



# Copyright 2017 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import logging
import os
import sys

from botocore.exceptions import ProfileNotFound
from botocore.history import get_global_history_recorder

from awscli.compat import sqlite3
from awscli.customizations.commands import BasicCommand
from awscli.customizations.history.constants import (
    DEFAULT_HISTORY_FILENAME,
    HISTORY_FILENAME_ENV_VAR,
)
from awscli.customizations.history.db import (
    DatabaseConnection,
    DatabaseHistoryHandler,
    DatabaseRecordWriter,
    RecordBuilder,
)
from awscli.customizations.history.list import ListCommand
from awscli.customizations.history.show import ShowCommand

LOG = logging.getLogger(__name__)
HISTORY_RECORDER = get_global_history_recorder()


def register_history_mode(event_handlers):
    event_handlers.register('session-initialized', attach_history_handler)


def register_history_commands(event_handlers):
    event_handlers.register(
        "building-command-table.main", add_history_commands
    )


def attach_history_handler(session, parsed_args, **kwargs):
    if _should_enable_cli_history(session, parsed_args):
        LOG.debug('Enabling CLI history')

        history_filename = os.environ.get(
            HISTORY_FILENAME_ENV_VAR, DEFAULT_HISTORY_FILENAME
        )
        if not os.path.isdir(os.path.dirname(history_filename)):
            os.makedirs(os.path.dirname(history_filename))

        try:
            connection = DatabaseConnection(history_filename)
        except Exception as e:
            LOG.debug('Unable to open history database: %s', e)
            sys.stderr.write(
                'Warning: Unable to record CLI history. '
                'Check file permissions for %s\n' % history_filename
            )
            return
        writer = DatabaseRecordWriter(connection)
        record_builder = RecordBuilder()
        db_handler = DatabaseHistoryHandler(writer, record_builder)

        HISTORY_RECORDER.add_handler(db_handler)
        HISTORY_RECORDER.enable()


def _should_enable_cli_history(session, parsed_args):
    if parsed_args.command == 'history':
        return False
    try:
        scoped_config = session.get_scoped_config()
    except ProfileNotFound:
        # If the profile does not exist, cli history is definitely not
        # enabled, but don't let the error get propagated as commands down
        # the road may handle this such as the configure set command with
        # a --profile flag set.
        return False
    has_history_enabled = scoped_config.get('cli_history') == 'enabled'
    if has_history_enabled and sqlite3 is None:
        if has_history_enabled:
            sys.stderr.write(
                'cli_history is enabled but sqlite3 is unavailable. '
                'Unable to collect CLI history.\n'
            )
        return False
    return has_history_enabled


def add_history_commands(command_table, session, **kwargs):
    command_table['history'] = HistoryCommand(session)


class HistoryCommand(BasicCommand):
    NAME = 'history'
    DESCRIPTION = (
        'Commands to interact with the history of AWS CLI commands ran '
        'over time. To record the history of AWS CLI commands set '
        '``cli_history`` to ``enabled`` in the ``~/.aws/config`` file. '
        'This can be done by running:\n\n'
        '``$ aws configure set cli_history enabled``'
    )
    SUBCOMMANDS = [
        {'name': 'show', 'command_class': ShowCommand},
        {'name': 'list', 'command_class': ListCommand},
    ]

    def _run_main(self, parsed_args, parsed_globals):
        if parsed_args.subcommand is None:
            self._raise_usage_error()



# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from awscli.compat import shlex

NOT_SET = '<not set>'
PREDEFINED_SECTION_NAMES = 'plugins'
# A map between the command line parameter name and the name used
# in the full config object.
SUBSECTION_TYPE_ALLOWLIST = {
    'sso-session': {"full_config_name": "sso_sessions"},
    'services': {"full_config_name": "services"},
}
_WHITESPACE = ' \t'


class ConfigValue:
    def __init__(self, value, config_type, config_variable):
        self.value = value
        self.config_type = config_type
        self.config_variable = config_variable

    def mask_value(self):
        if self.value is NOT_SET:
            return
        self.value = mask_value(self.value)


class SectionNotFoundError(Exception):
    pass


class SubsectionNotFoundError(Exception):
    pass


def mask_value(current_value):
    if current_value is None:
        return 'None'
    else:
        return ('*' * 16) + current_value[-4:]


def profile_to_section(profile_name):
    """Converts a profile name to a section header to be used in the config."""
    if profile_name == 'default':
        return profile_name
    return get_section_header('profile', profile_name)


def get_section_header(section_type, section_name):
    if any(c in _WHITESPACE for c in section_name):
        section_name = shlex.quote(section_name)
    return f'{section_type} {section_name}'



# Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import logging

import awscrt.io

LOG_FORMAT = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'


try:
    # Initializing CRT logging isn't thread-safe, but setting/updating it is.
    # So we always initialize logging when this module is imported so
    # subsequent enable/disable CRT logging calls can safely assume it's
    # already initialized.
    awscrt.io.init_logging(awscrt.io.LogLevel.NoLogs, 'stderr')
except RuntimeError:
    # Calling `init_logging` more than once raises a Runtime exception.
    # Even though normal usage shouldn't call it more than once in the
    # case of multiple imports, we should still guard against it if
    # something causes the module cache to be cleared between imports.
    pass


def set_stream_logger(logger_name, log_level, stream=None, format_string=None):
    """
    Convenience method to configure a stream logger.

    :type logger_name: str
    :param logger_name: The name of the logger to configure

    :type log_level: str
    :param log_level: The log level to set for the logger.  This
        is any param supported by the ``.setLevel()`` method of
        a ``Log`` object.

    :type stream: file
    :param stream: A file like object to log to.  If none is provided
        then sys.stderr will be used.

    :type format_string: str
    :param format_string: The format string to use for the log
        formatter.  If none is provided this will default to
        ``self.LOG_FORMAT``.

    """
    log = logging.getLogger(logger_name)
    log.setLevel(logging.DEBUG)

    remove_stream_logger(logger_name)

    handler_name = f'{logger_name}_stream_handler'
    ch = logging.StreamHandler(stream)
    ch.set_name(handler_name)
    ch.setLevel(log_level)

    # create formatter
    if format_string is None:
        format_string = LOG_FORMAT
    formatter = logging.Formatter(format_string)

    # add formatter to ch
    ch.setFormatter(formatter)

    # add ch to logger
    log.addHandler(ch)


def remove_stream_logger(logger_name):
    """
    Convenience method to configure a stream logger.

    :type logger_name: str
    :param logger_name: The name of the logger to configure

    """
    log = logging.getLogger(logger_name)
    handler_name = f'{logger_name}_stream_handler'
    for handler in log.handlers:
        if handler.get_name() == handler_name:
            log.removeHandler(handler)


def enable_crt_logging():
    awscrt.io.set_log_level(awscrt.io.LogLevel.Debug)


def disable_crt_logging():
    awscrt.io.set_log_level(awscrt.io.LogLevel.NoLogs)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

"""Color utilities for ECS Express Gateway Service monitoring."""

from colorama import Fore, Style, init  # noqa

# Status symbols
CHECK_MARK = '✓'


class ColorUtils:
    """Utility class for applying colors to text using colorama."""

    def __init__(self):
        # Initialize colorama
        init(autoreset=False, strip=False)

    def make_green(self, text, use_color=True):
        if not use_color:
            return text
        return f"{Fore.GREEN}{text}{Style.RESET_ALL}"

    def make_red(self, text, use_color=True):
        if not use_color:
            return text
        return f"{Fore.RED}{text}{Style.RESET_ALL}"

    def make_purple(self, text, use_color=True):
        if not use_color:
            return text
        return f"{Fore.MAGENTA}{text}{Style.RESET_ALL}"

    def make_yellow(self, text, use_color=True):
        if not use_color:
            return text
        return f"{Fore.YELLOW}{text}{Style.RESET_ALL}"

    def make_cyan(self, text, use_color=True):
        if not use_color:
            return text
        return f"{Fore.CYAN}{text}{Style.RESET_ALL}"

    def color_by_status(self, text, status, use_color=True):
        """Color text based on resource status."""
        if not status or status == "ACTIVE" or status == "SUCCESSFUL":
            return self.make_green(text, use_color)
        elif status == "FAILED":
            return self.make_red(text, use_color)
        elif status == "DELETED":
            return self.make_yellow(text, use_color)
        else:
            return self.make_purple(text, use_color)

    def make_status_symbol(self, status, spinner_char, use_color=True):
        """Create status symbol with appropriate color."""
        if not status or status == "ACTIVE" or status == "SUCCESSFUL":
            return self.make_green(f"{CHECK_MARK} ", use_color)
        elif status == "FAILED" or status == "ROLLBACK_FAILED":
            return self.make_red("X ", use_color)
        elif (
            status == "DELETED"
            or status == "STOPPED"
            or status == "ROLLBACK_SUCCESSFUL"
        ):
            return self.make_yellow("— ", use_color)
        else:
            return self.make_purple(f"{spinner_char} ", use_color)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""Generates auto completion index."""

import os

from awscli import clidriver
from awscli.autocomplete import db
from awscli.autocomplete.local import indexer
from awscli.autocomplete.serverside.indexer import APICallIndexer


def generate_index(filename):
    """Generates the default auto-complete index"""
    filename = os.path.abspath(filename)
    index_dir = os.path.dirname(filename)
    if not os.path.isdir(index_dir):
        os.makedirs(index_dir)

    # Using a temporary name so if the index already exists, we'll
    # only replace the entire file once we successfully regenerate the
    # index.
    temp_name = f'{filename}.temp'
    try:
        _do_generate_index(temp_name)
    except BaseException:
        if os.path.exists(temp_name):
            os.remove(temp_name)
        raise

    os.rename(temp_name, filename)
    return filename


def _do_generate_index(filename):
    db_connection = db.DatabaseConnection(filename)
    indexers = [
        indexer.ModelIndexer(db_connection),
        APICallIndexer(db_connection),
    ]
    driver = clidriver.create_clidriver()
    index_gen = IndexGenerator(indexers=indexers)
    try:
        index_gen.generate_index(driver)
    finally:
        db_connection.close()


class IndexGenerator:
    """Generates auto completion index.

    This will generate an auto completion index for all the low level
    indexer used by the CLI.  This object primarily delegates to other
    objects that do the actual heavy lifting of generating auto completion
    indices.

    """

    def __init__(self, indexers):
        self._indexers = indexers

    def generate_index(self, clidriver):
        for indexer in self._indexers:
            indexer.generate_index(clidriver)



# Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import json
import logging

from awscli.customizations.commands import BasicCommand
from awscli.customizations.emrcontainers.base36 import Base36
from awscli.customizations.emrcontainers.constants import (
    TRUST_POLICY_STATEMENT_ALREADY_EXISTS,
    TRUST_POLICY_STATEMENT_FORMAT,
    TRUST_POLICY_UPDATE_SUCCESSFUL,
)
from awscli.customizations.emrcontainers.eks import EKS
from awscli.customizations.emrcontainers.iam import IAM
from awscli.customizations.emrcontainers.utils import get_region
from awscli.customizations.utils import get_policy_arn_suffix, uni_print

LOG = logging.getLogger(__name__)


def check_if_statement_exists(expected_statement, actual_assume_role_document):
    if actual_assume_role_document is None:
        return False

    existing_statements = actual_assume_role_document.get("Statement", [])
    for existing_statement in existing_statements:
        matches = check_if_dict_matches(expected_statement, existing_statement)
        if matches:
            return True

    return False


def check_if_dict_matches(expected_dict, actual_dict):
    if len(expected_dict) != len(actual_dict):
        return False

    for key in expected_dict:
        key_str = str(key)
        val = expected_dict[key_str]
        if isinstance(val, dict):
            if not check_if_dict_matches(val, actual_dict.get(key_str, {})):
                return False
        else:
            if key_str not in actual_dict or actual_dict[key_str] != str(val):
                return False

    return True


class UpdateRoleTrustPolicyCommand(BasicCommand):
    NAME = 'update-role-trust-policy'

    DESCRIPTION = BasicCommand.FROM_FILE(
        'emr-containers', 'update-role-trust-policy', '_description.rst'
    )

    ARG_TABLE = [
        {
            'name': 'cluster-name',
            'help_text': (
                "Specify the name of the Amazon EKS cluster with "
                "which the IAM Role would be used."
            ),
            'required': True,
        },
        {
            'name': 'namespace',
            'help_text': (
                "Specify the namespace from the Amazon EKS cluster "
                "with which the IAM Role would be used."
            ),
            'required': True,
        },
        {
            'name': 'role-name',
            'help_text': (
                "Specify the IAM Role name that you want to use"
                "with Amazon EMR on EKS."
            ),
            'required': True,
        },
        {
            'name': 'iam-endpoint',
            'no_paramfile': True,
            'help_text': (
                "The  IAM  endpoint  to call for updating the role "
                "trust policy. This is optional and should only be"
                "specified when a custom endpoint should be called"
                "for IAM operations."
            ),
            'required': False,
        },
        {
            'name': 'dry-run',
            'action': 'store_true',
            'default': False,
            'help_text': (
                "Print the merged trust policy document to"
                "stdout instead of updating the role trust"
                "policy directly."
            ),
            'required': False,
        },
    ]

    def _run_main(self, parsed_args, parsed_globals):
        """Call to run the commands"""

        self._cluster_name = parsed_args.cluster_name
        self._namespace = parsed_args.namespace
        self._role_name = parsed_args.role_name
        self._region = get_region(self._session, parsed_globals)
        self._endpoint_url = parsed_args.iam_endpoint
        self._dry_run = parsed_args.dry_run

        result = self._update_role_trust_policy(parsed_globals)
        uni_print(result)
        uni_print("\n")

        return 0

    def _update_role_trust_policy(self, parsed_globals):
        """Method to update  trust policy if not done already"""

        base36 = Base36()

        eks_client = EKS(
            self._session.create_client(
                'eks',
                region_name=self._region,
                verify=parsed_globals.verify_ssl,
            )
        )

        account_id = eks_client.get_account_id(self._cluster_name)
        oidc_provider = eks_client.get_oidc_issuer_id(self._cluster_name)

        base36_encoded_role_name = base36.encode(self._role_name)
        LOG.debug('Base36 encoded role name: %s', base36_encoded_role_name)
        trust_policy_statement = json.loads(
            TRUST_POLICY_STATEMENT_FORMAT
            % {
                "AWS_ACCOUNT_ID": account_id,
                "OIDC_PROVIDER": oidc_provider,
                "NAMESPACE": self._namespace,
                "BASE36_ENCODED_ROLE_NAME": base36_encoded_role_name,
                "AWS_PARTITION": get_policy_arn_suffix(self._region),
            }
        )

        LOG.debug(
            'Computed Trust Policy Statement:\n%s',
            json.dumps(trust_policy_statement, indent=2),
        )
        iam_client = IAM(
            self._session.create_client(
                'iam',
                region_name=self._region,
                endpoint_url=self._endpoint_url,
                verify=parsed_globals.verify_ssl,
            )
        )

        assume_role_document = iam_client.get_assume_role_policy(
            self._role_name
        )
        matches = check_if_statement_exists(
            trust_policy_statement, assume_role_document
        )

        if not matches:
            LOG.debug(
                'Role %s does not have the required trust policy ',
                self._role_name,
            )

            existing_statements = assume_role_document.get("Statement")
            if existing_statements is None:
                assume_role_document["Statement"] = [trust_policy_statement]
            else:
                existing_statements.append(trust_policy_statement)

            if self._dry_run:
                return json.dumps(assume_role_document, indent=2)
            else:
                LOG.debug('Updating trust policy of role %s', self._role_name)
                iam_client.update_assume_role_policy(
                    self._role_name, assume_role_document
                )
                return TRUST_POLICY_UPDATE_SUCCESSFUL % self._role_name
        else:
            return TRUST_POLICY_STATEMENT_ALREADY_EXISTS % self._role_name



# Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import os

from prompt_toolkit.application import get_app
from prompt_toolkit.buffer import Buffer
from prompt_toolkit.completion import PathCompleter
from prompt_toolkit.filters import Condition, has_focus
from prompt_toolkit.formatted_text import HTML, to_formatted_text
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.key_binding.bindings.focus import (
    focus_next,
    focus_previous,
)
from prompt_toolkit.keys import Keys
from prompt_toolkit.layout import Layout
from prompt_toolkit.layout.containers import (
    ConditionalContainer,
    Dimension,
    HSplit,
    VSplit,
    Window,
    WindowAlign,
    to_container,
    to_filter,
)
from prompt_toolkit.layout.controls import BufferControl, FormattedTextControl
from prompt_toolkit.utils import is_windows
from prompt_toolkit.widgets import (
    Box,
    Button,
    Dialog,
    Frame,
    HorizontalLine,
    Label,
    Shadow,
    TextArea,
    VerticalLine,
)

from awscli.autoprompt.widgets import BaseToolbarView, TitleLine
from awscli.customizations.wizard import core
from awscli.customizations.wizard.ui.keybindings import (
    details_visible,
    error_bar_enabled,
    prompt_has_details,
    save_details_visible,
)
from awscli.customizations.wizard.ui.section import (
    WizardSectionBody,
    WizardSectionTab,
)
from awscli.customizations.wizard.ui.utils import (
    Spacer,
    get_ui_control_by_buffer_name,
    move_to_previous_prompt,
)


class WizardLayoutFactory:
    def create_wizard_layout(self, definition):
        run_wizard_dialog = self._create_run_wizard_dialog(definition)
        error_bar = self._create_error_bar()
        container = HSplit(
            [
                self._create_title(definition),
                self._create_sections(
                    definition, run_wizard_dialog, error_bar
                ),
                HorizontalLine(),
            ]
        )
        return WizardLayout(
            container=container,
            run_wizard_dialog=run_wizard_dialog,
            error_bar=error_bar,
        )

    def _create_title(self, definition):
        title = Label(f'{definition["title"]}', style='class:wizard.title')
        title.window.align = WindowAlign.CENTER
        return title

    def _create_sections(self, definition, run_wizard_dialog, error_bar):
        section_tabs = []
        section_bodies = []
        for section_name, section_definition in definition['plan'].items():
            if section_name == core.DONE_SECTION_NAME:
                section_tabs.append(
                    self._create_done_section_tab(section_definition)
                )
                section_bodies.append(run_wizard_dialog)
            else:
                section_tabs.append(
                    self._create_section_tab(section_name, section_definition)
                )
                section_bodies.append(
                    self._create_section_body(section_name, section_definition)
                )
        section_tabs.append(Spacer())
        return VSplit(
            [
                HSplit(
                    section_tabs, padding=1, style='class:wizard.section.tab'
                ),
                ConditionalContainer(
                    VerticalLine(), filter=Condition(is_windows)
                ),
                HSplit(
                    [
                        *section_bodies,
                        WizardDetailsPanel(),
                        error_bar,
                        ToolbarView(),
                    ]
                ),
            ]
        )

    def _create_section_tab(self, section_name, section_definition):
        return WizardSectionTab(section_name, section_definition)

    def _create_done_section_tab(self, done_definition):
        if not done_definition or 'shortname' not in done_definition:
            done_definition = {'shortname': 'Done'}
        return WizardSectionTab(core.DONE_SECTION_NAME, done_definition)

    def _create_section_body(self, section_name, section_definition):
        return WizardSectionBody(section_name, section_definition)

    def _create_run_wizard_dialog(self, wizard_defintion):
        done_section = wizard_defintion['plan'][core.DONE_SECTION_NAME]
        return RunWizardDialog.from_done_section_definition(done_section)

    def _create_error_bar(self):
        return WizardErrorBar()


class WizardLayout(Layout):
    def __init__(self, container, run_wizard_dialog, error_bar):
        self.run_wizard_dialog = run_wizard_dialog
        self.error_bar = error_bar
        super().__init__(container)


class WizardDetailsPanel:
    DIMENSIONS = {
        'details_window_height_max': 40,
        'details_window_height_pref': 30,
    }

    def __init__(self):
        self.container = self._get_container()

    def _get_title(self):
        return getattr(get_app(), 'details_title', '') or "Details panel"

    def _get_container(self):
        return ConditionalContainer(
            HSplit(
                [
                    TitleLine(self._get_title),
                    VSplit(
                        [
                            Window(
                                content=BufferControl(
                                    buffer=Buffer(
                                        name='details_buffer', read_only=True
                                    ),
                                ),
                                height=Dimension(
                                    max=self.DIMENSIONS[
                                        'details_window_height_max'
                                    ],
                                    preferred=self.DIMENSIONS[
                                        'details_window_height_pref'
                                    ],
                                ),
                                wrap_lines=True,
                            ),
                            SaveFileDialogue(),
                        ]
                    ),
                ]
            ),
            details_visible,
        )

    def __pt_container__(self):
        return self.container


class SaveFileDialogue:
    def __init__(self):
        self.container = self._get_container()

    def _get_container(self):
        return ConditionalContainer(
            self._create_dialog(), save_details_visible
        )

    def __pt_container__(self):
        return self.container

    def _create_dialog(self):
        textfield = self._create_textfield()
        save_button = self._create_save_button(textfield)
        cancel_button = self._create_cancel_button(textfield)
        dialog = Dialog(
            title='Save to file',
            body=HSplit(
                [
                    Label(text='Filename', dont_extend_height=True),
                    textfield,
                ],
                padding=Dimension(preferred=1, max=1),
            ),
            buttons=[save_button, cancel_button],
            with_background=True,
        )
        dialog.container.container.style = 'class:wizard.dialog.save'
        dialog.container.body.container.style = 'class:wizard.dialog.body'
        dialog.container.body.container.content.key_bindings.add(
            Keys.Enter, filter=has_focus('save_details_dialogue')
        )(save_button.handler)
        return dialog

    def _create_textfield(self):
        text_area = TextArea(
            multiline=False,
            completer=PathCompleter(),
            complete_while_typing=True,
        )
        text_area.buffer.name = 'save_details_dialogue'
        return text_area

    def _create_save_button(self, textfield):
        def save_handler(*args, **kwargs):
            app = get_app()
            contents = app.layout.get_buffer_by_name(
                'details_buffer'
            ).document.text
            app.file_io.write_file_contents(textfield.text, contents)
            app.save_details_visible = False
            current_control = get_ui_control_by_buffer_name(
                app.layout, app.traverser.get_current_prompt()
            )
            app.layout.focus(current_control)

        return Button(text='Save', handler=save_handler)

    def _create_cancel_button(self, textfield):
        def cancel_handler(*args, **kwargs):
            app = get_app()
            app.save_details_visible = False
            app.layout.focus('details_buffer')

        return Button(text='Cancel', handler=cancel_handler)

    def _create_key_bindings(self, save_button, cancel_button):
        kb = KeyBindings()
        first_selected = has_focus(save_button)
        last_selected = has_focus(cancel_button)

        kb.add(Keys.Left, filter=~first_selected)(focus_previous)
        kb.add(Keys.Right, filter=~last_selected)(focus_next)
        kb.add(Keys.Tab)(focus_next)
        kb.add(Keys.BackTab)(focus_previous)
        return kb


class ToolbarView(BaseToolbarView):
    CONDITION = prompt_has_details | error_bar_enabled

    def __init__(self):
        self.content = to_container(self.create_window(self.help_text))
        self.filter = to_filter(self.CONDITION)

    def create_window(self, help_text):
        text_control = FormattedTextControl(text=lambda: help_text)
        text_control.name = 'toolbar_panel'
        return HSplit(
            [
                HorizontalLine(),
                Window(
                    content=text_control, wrap_lines=True, **self.DIMENSIONS
                ),
            ]
        )

    def help_text(self):
        app = get_app()
        output = []
        if prompt_has_details():
            title = getattr(app, 'details_title', 'Details panel')
            output.extend(
                [
                    f'{self.STYLE}[F2]</style> Switch to {title}',
                    f'{self.STYLE}[F3]</style> Show/Hide {title}',
                    f'{self.STYLE}[CTRL+S]</style> Save {title}',
                ]
            )
        if error_bar_enabled():
            output.append(f'{self.STYLE}[F4]</style> Show/Hide error message')
        return to_formatted_text(HTML(f'{self.SPACING}'.join(output)))


class RunWizardDialogButtonFactory:
    def create_button(self, button_name, text=None):
        button_kwargs = {}
        if text:
            button_kwargs['text'] = text
        return getattr(self, f'_create_{button_name}_button')(**button_kwargs)

    def _create_yes_button(self, text='Yes'):
        def yes_handler():
            app = get_app()
            try:
                app.traverser.run_wizard()
                app.exit(result=0)
            except Exception as e:
                app.layout.error_bar.display_error(e)
                move_to_previous_prompt(app)

        return self._create_dialog_button(text=text, handler=yes_handler)

    def _create_back_button(self, text='Back'):
        def back_handler():
            app = get_app()
            app.layout.error_bar.clear()
            move_to_previous_prompt(app)

        return self._create_dialog_button(text=text, handler=back_handler)

    def _create_dialog_button(self, text, handler):
        button = Button(text=text, handler=handler)
        button.window.always_hide_cursor = to_filter(True)
        return button


class RunWizardDialog:
    _DEFAULT_BUTTONS = ['yes', 'back']

    def __init__(self, title='Run wizard?', buttons=None):
        self._title = title
        if buttons is None:
            buttons = self._create_default_buttons()
        self._buttons = buttons
        self.container = self._get_container()

    @classmethod
    def from_done_section_definition(cls, done_section_definition):
        cls_kwargs = {}
        if done_section_definition:
            if 'description' in done_section_definition:
                cls_kwargs['title'] = done_section_definition['description']
            if 'options' in done_section_definition:
                buttons = []
                button_factory = RunWizardDialogButtonFactory()
                for option in done_section_definition['options']:
                    create_button_kwargs = {'button_name': option}
                    if isinstance(option, dict):
                        create_button_kwargs['button_name'] = option['name']
                        create_button_kwargs['text'] = option.get(
                            'description'
                        )
                    buttons.append(
                        button_factory.create_button(**create_button_kwargs)
                    )
                cls_kwargs['buttons'] = buttons
        return cls(**cls_kwargs)

    def _get_container(self):
        dialog = Box(
            body=self._create_dialog_frame(),
            style='class:wizard.dialog',
            padding_top=5,
        )
        return ConditionalContainer(
            dialog,
            Condition(self._is_visible),
        )

    def _is_visible(self):
        return get_app().traverser.has_no_remaining_prompts()

    def _create_default_buttons(self):
        buttons = []
        button_factory = RunWizardDialogButtonFactory()
        for button_name in self._DEFAULT_BUTTONS:
            buttons.append(button_factory.create_button(button_name))
        return buttons

    def _create_dialog_frame(self):
        frame_body = Box(
            body=self._create_buttons_container(),
            height=Dimension(min=1, max=3, preferred=3),
        )
        return Shadow(
            body=Frame(
                title=self._title,
                body=frame_body,
                style='class:wizard.dialog.body',
            )
        )

    def _create_buttons_container(self):
        buttons_kb = self._create_buttons_key_bindings(self._buttons)
        return VSplit(self._buttons, padding=1, key_bindings=buttons_kb)

    def _create_buttons_key_bindings(self, buttons):
        kb = KeyBindings()
        first_selected = has_focus(buttons[0])
        last_selected = has_focus(buttons[-1])

        kb.add(Keys.Left, filter=~first_selected)(focus_previous)
        kb.add(Keys.Right, filter=~last_selected)(focus_next)
        kb.add(Keys.Tab)(focus_next)
        kb.add(Keys.BackTab)(focus_previous)
        return kb

    def __pt_container__(self):
        return self.container


class WizardErrorBar:
    def __init__(self):
        self._error_bar_buffer = self._get_error_bar_buffer()
        self.container = self._get_container()
        self.current_error = None

    def display_error(self, exception):
        self.current_error = exception
        self._error_bar_buffer.text = (
            f'Encountered following error in wizard:\n\n{exception}'
        )
        get_app().error_bar_visible = True

    def clear(self):
        self.current_error = None
        self._error_bar_buffer.text = ''
        get_app().error_bar_visible = None

    def _get_error_bar_buffer(self):
        return Buffer(name='error_bar')

    def _get_container(self):
        return ConditionalContainer(
            HSplit(
                [
                    TitleLine('Wizard exception'),
                    Window(
                        content=BufferControl(
                            buffer=self._error_bar_buffer, focusable=False
                        ),
                        style='class:wizard.error',
                        dont_extend_height=True,
                        wrap_lines=True,
                    ),
                ]
            ),
            Condition(self._is_visible),
        )

    def _is_visible(self):
        return get_app().error_bar_visible

    def __pt_container__(self):
        return self.container



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import codecs
import os
import sys

from awscli.customizations.commands import BasicCommand
from awscli.customizations.configure.writer import ConfigFileWriter
from awscli.customizations.utils import uni_print


class ConfigureImportCommand(BasicCommand):
    NAME = 'import'
    DESCRIPTION = (
        'Import CSV credentials generated from the AWS web console. '
        'Entries in the CSV will be imported as profiles in the AWS '
        'credentials file, with the profile name matching the IAM User Name.'
    )
    EXAMPLES = (
        'aws configure import --csv file://credentials.csv\n\n'
        'aws configure import --csv file://credentials.csv --skip-invalid\n\n'
        'aws configure import --csv file://credentials.csv '
        '--profile-prefix test-\n\n'
    )
    ARG_TABLE = [
        {
            'name': 'csv',
            'required': True,
            'help_text': (
                'The credentials in CSV format generated by the AWS web console. '
                'The CSV file must contain the "User name", "Access key ID", and '
                '"Secret access key" headers.'
                'If passing a CSV file path instead of a CSV-formatted string, '
                'the "file://" prefix is required.'
            ),
            'cli_type_name': 'string',
        },
        {
            'name': 'skip-invalid',
            'dest': 'skip_invalid',
            'help_text': (
                'Skip entries that are invalid or do not have programmatic '
                'access instead of failing.'
            ),
            'default': False,
            'action': 'store_true',
        },
        {
            'name': 'profile-prefix',
            'dest': 'profile_prefix',
            'help_text': (
                'Adds the specified prefix to the beginning of all profile names.'
            ),
            'default': '',
            'cli_type_name': 'string',
        },
    ]

    def __init__(
        self, session, csv_parser=None, importer=None, out_stream=None
    ):
        super(ConfigureImportCommand, self).__init__(session)
        if csv_parser is None:
            csv_parser = CSVCredentialParser()
        self._csv_parser = csv_parser

        if importer is None:
            writer = ConfigFileWriter()
            importer = CredentialImporter(writer)
        self._importer = importer

        if out_stream is None:
            out_stream = sys.stdout
        self._out_stream = out_stream

    def _get_config_path(self):
        config_file = self._session.get_config_variable('credentials_file')
        return os.path.expanduser(config_file)

    def _import_csv(self, contents):
        self._check_possible_filepath(contents)
        config_path = self._get_config_path()
        credentials = self._csv_parser.parse_credentials(contents)
        for credential in credentials:
            self._importer.import_credential(
                credential,
                config_path,
                profile_prefix=self._profile_prefix,
            )
        import_msg = 'Successfully imported %s profile(s)\n' % len(credentials)
        uni_print(import_msg, out_file=self._out_stream)

    def _check_possible_filepath(self, csv_data):
        if ('\n' not in csv_data and
            os.path.exists(csv_data) and
            not csv_data.startswith('file://')):
            raise ValueError(
                "You may be passing a file to import without the 'file://' prefix. "
                "To import a CSV file, use --csv file://path/to/file.csv"
            )

    def _run_main(self, parsed_args, parsed_globals):
        self._csv_parser.strict = not parsed_args.skip_invalid
        self._profile_prefix = parsed_args.profile_prefix
        self._import_csv(parsed_args.csv)
        return 0


class CredentialParserError(Exception):
    pass


class CSVCredentialParser:
    _USERNAME_HEADER = 'User Name'
    _AKID_HEADER = 'Access Key ID'
    _SAK_HEADER = 'Secret Access key'
    _EXPECTED_HEADERS = [_USERNAME_HEADER, _AKID_HEADER, _SAK_HEADER]

    _EMPTY_CSV = 'Provided CSV contains no contents'
    _HEADER_NOT_FOUND = 'Expected header "%s" not found'
    _ROW_MISSING_HEADER = 'Row missing value for header "%s"'
    _INVALID_ROW = 'Failed to parse entry #%s: %s'

    def __init__(self, strict=True):
        self.strict = strict

    def _format_header(self, header):
        # Remove leading UTF BOM character if present
        if header.startswith(codecs.BOM_UTF8.decode()):
            header = header[1:]
        return header.lower().strip()

    def _parse_csv_headers(self, header):
        return [self._format_header(h) for h in header.split(',')]

    def _extract_expected_header_indices(self, headers):
        indices = {}
        for header in self._EXPECTED_HEADERS:
            formatted_header = self._format_header(header)
            if formatted_header not in headers:
                raise CredentialParserError(self._HEADER_NOT_FOUND % header)
            indices[header] = headers.index(formatted_header)
        return indices

    def _parse_csv_row(self, row, header_indices):
        item = {}
        cols = row.split(',')
        for header, index in header_indices.items():
            try:
                item[header] = cols[index].strip()
            except IndexError:
                item[header] = None
            if not item[header]:
                raise CredentialParserError(self._ROW_MISSING_HEADER % header)
        return item

    def _parse_csv_rows(self, rows, header_indices):
        count = 0
        parsed_rows = []
        for row in rows:
            count += 1
            try:
                item = self._parse_csv_row(row, header_indices)
            except CredentialParserError as e:
                if not self.strict:
                    continue
                raise CredentialParserError(self._INVALID_ROW % (count, e))
            parsed_rows.append(item)
        return parsed_rows

    def _parse_csv(self, csv):
        if not csv.strip():
            raise CredentialParserError(self._EMPTY_CSV)

        lines = csv.splitlines()
        parsed_headers = self._parse_csv_headers(lines[0])
        header_indices = self._extract_expected_header_indices(parsed_headers)
        return self._parse_csv_rows(lines[1:], header_indices)

    def _convert_rows_to_credentials(self, parsed_rows):
        credentials = []
        for row in parsed_rows:
            username = row.get(self._USERNAME_HEADER)
            akid = row.get(self._AKID_HEADER)
            sak = row.get(self._SAK_HEADER)
            credentials.append((username, akid, sak))
        return credentials

    def parse_credentials(self, contents):
        # Expected format is:
        # User name,Password,Access key ID,Secret access key,Console login link
        # username1,pw,akid,sak,https://console.link
        # username2,pw,akid,sak,https://console.link
        parsed_rows = self._parse_csv(contents)
        return self._convert_rows_to_credentials(parsed_rows)


class CredentialImporter:
    def __init__(self, writer):
        self._config_writer = writer

    def import_credential(
        self, credential, credentials_file, profile_prefix=''
    ):
        name, akid, sak = credential
        config_profile = {
            '__section__': profile_prefix + name,
            'aws_access_key_id': akid,
            'aws_secret_access_key': sak,
        }
        self._config_writer.update_config(config_profile, credentials_file)



# Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import json
import os
from asyncio import new_event_loop, set_event_loop
from collections.abc import MutableMapping

from prompt_toolkit.application import Application

from awscli.customizations.wizard import core
from awscli.customizations.wizard.exceptions import (
    InvalidChoiceException,
    UnableToRunWizardError,
    UnexpectedWizardException,
)
from awscli.customizations.wizard.ui.keybindings import get_default_keybindings
from awscli.customizations.wizard.ui.style import get_default_style
from awscli.utils import json_encoder


class WizardAppRunner:
    def __init__(self, session, app_factory):
        self._session = session
        self._app_factory = app_factory

    def run(self, loaded):
        """Run a single wizard given the contents as a string."""
        app = self._app_factory(loaded, self._session)
        app.run()
        print(app.traverser.get_output())


class WizardApp(Application):
    def __init__(
        self,
        layout,
        values,
        traverser,
        executor,
        style=None,
        key_bindings=None,
        full_screen=True,
        output=None,
        app_input=None,
        file_io=None,
    ):
        self.values = values
        self.traverser = traverser
        self.executor = executor
        if style is None:
            style = get_default_style()
        if key_bindings is None:
            key_bindings = get_default_keybindings()
        self.details_visible = False
        self.error_bar_visible = None
        self.save_details_visible = False
        if file_io is None:
            file_io = FileIO()
        self.file_io = file_io
        super().__init__(
            layout=layout,
            style=style,
            key_bindings=key_bindings,
            full_screen=full_screen,
            output=output,
            input=app_input,
        )

    def run(self, pre_run=None, **kwargs):
        loop = new_event_loop()
        try:
            set_event_loop(loop)
            loop.set_exception_handler(self._handle_exception)
            f = self.run_async(pre_run=pre_run, set_exception_handler=False)
            return loop.run_until_complete(f)
        finally:
            loop.close()

    def _handle_exception(self, loop, context):
        self.exit(exception=UnexpectedWizardException(context['exception']))


class WizardTraverser:
    DONE = core.DONE_SECTION_NAME
    OUTPUT = core.OUTPUT_SECTION_NAME

    def __init__(self, definition, values, executor):
        self._definition = definition
        self._values = values
        self._executor = executor
        self._prompt_definitions = self._collect_prompt_definitions()
        self._prompt_to_sections = self._map_prompts_to_sections()
        self._current_prompt = self._get_starting_prompt()
        self._previous_prompts = []
        self._visited_sections = [self.get_current_section()]

    def get_current_prompt(self):
        return self._current_prompt

    def get_current_section(self):
        if self.has_no_remaining_prompts():
            return self.DONE
        return self._prompt_to_sections[self._current_prompt]

    def get_current_prompt_choices(self):
        choices = self._get_choices(self._current_prompt)
        if choices:
            return [choice['display'] for choice in choices]
        return None

    def current_prompt_has_details(self):
        return 'details' in self._prompt_definitions.get(
            self._current_prompt, {}
        )

    def submit_prompt_answer(self, answer):
        definition = self._prompt_definitions[self._current_prompt]
        if 'choices' in definition:
            answer = self._convert_display_value_to_actual_value(
                self._get_choices(self._current_prompt), answer
            )
        if 'datatype' in definition:
            answer = core.DataTypeConverter.convert(
                definition['datatype'], answer
            )

        self._values[self._current_prompt] = answer

    def next_prompt(self):
        if self._current_prompt == self.DONE:
            return self._current_prompt
        new_prompt = self._get_next_prompt()
        self._previous_prompts.append(self._current_prompt)
        section_of_new_prompt = self._prompt_to_sections.get(new_prompt)
        if section_of_new_prompt != self.get_current_section():
            self._visited_sections.append(self._prompt_to_sections[new_prompt])
        self._current_prompt = new_prompt
        return new_prompt

    def previous_prompt(self):
        if self._previous_prompts:
            previous = self._previous_prompts.pop()
            self._current_prompt = previous
        return self._current_prompt

    def run_wizard(self):
        if self.has_no_remaining_prompts():
            self._executor.execute(self._definition['execute'], self._values)
        else:
            raise UnableToRunWizardError(
                'Cannot run wizard. There are prompts remaining that must '
                'be answered.'
            )

    def is_prompt_visible(self, value_name):
        if self._prompt_to_sections[value_name] != self.get_current_section():
            return False
        return self._prompt_meets_condition(value_name)

    def is_prompt_details_visible_by_default(self, value_name):
        return (
            self._prompt_definitions[value_name]
            .get('details', {})
            .get('visible', False)
        )

    def has_visited_section(self, section_name):
        return section_name in self._visited_sections

    def has_no_remaining_prompts(self):
        return self.get_current_prompt() == self.DONE

    def _collect_prompt_definitions(self):
        value_prompt_definitions = {}
        plan = self._definition['plan']
        for section_name, section_definition in plan.items():
            if section_name == self.DONE:
                continue
            for name, value_definition in section_definition['values'].items():
                if value_definition['type'] == 'prompt':
                    value_prompt_definitions[name] = value_definition
        return value_prompt_definitions

    def _map_prompts_to_sections(self):
        prompts_to_sections = {}
        sections = self._definition['plan']
        for section_name, section_definition in sections.items():
            if section_name == self.DONE:
                prompts_to_sections[section_name] = self.DONE
                continue
            for name, value_definition in section_definition['values'].items():
                prompts_to_sections[name] = section_name
        return prompts_to_sections

    def _get_starting_prompt(self):
        return list(self._prompt_definitions)[0]

    def _get_choices(self, value_name):
        value_definition = self._prompt_definitions.get(value_name, {})
        if 'choices' in value_definition:
            choices = value_definition['choices']
            if not isinstance(value_definition['choices'], list):
                # When choices is declared as a variable, we want to use the
                # value from the variable.
                choices = self._values[value_definition['choices']]
            if choices:
                return self._get_normalized_choice_values(choices)
        return None

    def get_details_title(self):
        step_definition = self._prompt_definitions[self._current_prompt]
        if 'details' in step_definition:
            return step_definition['details'].get('description')
        return ''

    def get_details_for_choice(self, choice):
        if not choice:
            return ''
        step_definition = self._prompt_definitions[self._current_prompt]
        if 'details' in step_definition:
            value = self._convert_display_value_to_actual_value(
                self._get_choices(self._current_prompt), choice
            )
            format_type = step_definition['details'].get('output', 'text')
            temp_values = self._values.copy()
            if step_definition['details']['value'] == '__selected_choice__':
                output = temp_values[value]
            else:
                temp_values[self._current_prompt] = value
                output = temp_values[step_definition['details']['value']]

            return self._format_output(output, format_type)

    def _format_output(self, output, format_type):
        if format_type == 'json':
            return json.dumps(output, indent=4, default=json_encoder)
        return output

    def _get_normalized_choice_values(self, choices):
        normalized_choices = []
        for choice in choices:
            if isinstance(choice, str):
                normalized_choices.append(
                    {'display': choice, 'actual_value': choice}
                )
            else:
                normalized_choices.append(choice)
        return normalized_choices

    def _convert_display_value_to_actual_value(self, choices, display_value):
        if choices:
            for choice in choices:
                if choice['display'] == display_value:
                    return choice['actual_value']
            raise InvalidChoiceException(
                f"'{display_value}' is not a valid choice. Valid choices are: "
                f"{[choice['display'] for choice in choices]}"
            )

    def _get_next_prompt(self):
        prompts = list(self._prompt_definitions)
        current_pos = prompts.index(self._current_prompt)
        for prompt in prompts[current_pos + 1 :]:
            if self._prompt_meets_condition(prompt):
                return prompt
        return self.DONE

    def _prompt_meets_condition(self, value_name):
        value_definition = self._prompt_definitions[value_name]
        if 'condition' in value_definition:
            return core.ConditionEvaluator().evaluate(
                value_definition['condition'], self._values
            )
        return True

    def get_output(self):
        template_step = core.TemplateStep()
        return template_step.run_step(
            self._definition[self.OUTPUT], self._values
        )


class WizardValues(MutableMapping):
    def __init__(
        self, definition, value_retrieval_steps=None, exception_handler=None
    ):
        self._definition = definition
        if value_retrieval_steps is None:
            value_retrieval_steps = {}
        self._value_retrieval_steps = value_retrieval_steps
        self._values = {}
        self._value_definitions = self._collect_all_value_definitions()
        self._exception_handler = exception_handler

    def __getitem__(self, key):
        if key not in self._values and key in self._value_definitions:
            value_definition = self._value_definitions[key]
            if value_definition['type'] in self._value_retrieval_steps:
                retrieval_step = self._value_retrieval_steps[
                    value_definition['type']
                ]
                try:
                    return retrieval_step.run_step(value_definition, self)
                except Exception as e:
                    if self._exception_handler is not None:
                        return self._exception_handler(e)
                    raise
        return self._values[key]

    def __setitem__(self, key, value):
        self._values[key] = value

    def __delitem__(self, key):
        del self._values[key]

    def __iter__(self):
        return iter(self._values)

    def __len__(self):
        return len(self._values)

    def __str__(self):
        return str(self._values)

    def _collect_all_value_definitions(self):
        value_definitions = {}
        plan = self._definition['plan']
        for section_name, section_definition in plan.items():
            if section_name == core.DONE_SECTION_NAME:
                continue
            for name, value_definition in section_definition['values'].items():
                value_definitions[name] = value_definition
        return value_definitions

    def __copy__(self):
        new_vars = type(self)(
            self._definition,
            self._value_retrieval_steps,
        )
        for varname, value in self.items():
            new_vars[varname] = value
        return new_vars

    copy = __copy__


class FileIO:
    def write_file_contents(self, filename, contents):
        with open(os.path.expanduser(filename), 'w') as f:
            f.write(contents)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from awscli.customizations.eks.get_token import GetTokenCommand
from awscli.customizations.eks.update_kubeconfig import UpdateKubeconfigCommand


def initialize(cli):
    """
    The entry point for EKS high level commands.
    """
    cli.register('building-command-table.eks', inject_commands)


def inject_commands(command_table, session, **kwargs):
    """
    Called when the EKS command table is being built.
    Used to inject new high level commands into the command list.
    """
    command_table['update-kubeconfig'] = UpdateKubeconfigCommand(session)
    command_table['get-token'] = GetTokenCommand(session)



# Copyright 2012-2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import logging
import os
import sys
import threading

import botocore
import botocore.exceptions
from botocore.compat import get_md5
from s3transfer.manager import TransferManager
from s3transfer.subscribers import BaseSubscriber

from awscli.compat import collections_abc

LOG = logging.getLogger(__name__)


def _get_checksum():
    hashlib_params = {"usedforsecurity": False}
    try:
        checksum = get_md5(**hashlib_params)
    except botocore.exceptions.MD5UnavailableError:
        import hashlib
        checksum = hashlib.sha256(**hashlib_params)
    return checksum


class NoSuchBucketError(Exception):
    def __init__(self, **kwargs):
        msg = self.fmt.format(**kwargs)
        Exception.__init__(self, msg)
        self.kwargs = kwargs

    fmt = (
        "S3 Bucket does not exist. "
        "Execute the command to create a new bucket"
        "\n"
        "aws s3 mb s3://{bucket_name}"
    )


class S3Uploader:
    """
    Class to upload objects to S3 bucket that use versioning. If bucket
    does not already use versioning, this class will turn on versioning.
    """

    @property
    def artifact_metadata(self):
        """
        Metadata to attach to the object(s) uploaded by the uploader.
        """
        return self._artifact_metadata

    @artifact_metadata.setter
    def artifact_metadata(self, val):
        if val is not None and not isinstance(val, collections_abc.Mapping):
            raise TypeError("Artifact metadata should be in dict type")
        self._artifact_metadata = val

    def __init__(
        self,
        s3_client,
        bucket_name,
        prefix=None,
        kms_key_id=None,
        force_upload=False,
        transfer_manager=None,
    ):
        self.bucket_name = bucket_name
        self.prefix = prefix
        self.kms_key_id = kms_key_id or None
        self.force_upload = force_upload
        self.s3 = s3_client

        self.transfer_manager = transfer_manager
        if not transfer_manager:
            self.transfer_manager = TransferManager(self.s3)

        self._artifact_metadata = None

    def upload(self, file_name, remote_path):
        """
        Uploads given file to S3
        :param file_name: Path to the file that will be uploaded
        :param remote_path:  be uploaded
        :return: VersionId of the latest upload
        """

        if self.prefix and len(self.prefix) > 0:
            remote_path = f"{self.prefix}/{remote_path}"

        # Check if a file with same data exists
        if not self.force_upload and self.file_exists(remote_path):
            LOG.debug(
                f"File with same data already exists at {remote_path}. "
                "Skipping upload"
            )
            return self.make_url(remote_path)

        try:
            # Default to regular server-side encryption unless customer has
            # specified their own KMS keys
            additional_args = {"ServerSideEncryption": "AES256"}

            if self.kms_key_id:
                additional_args["ServerSideEncryption"] = "aws:kms"
                additional_args["SSEKMSKeyId"] = self.kms_key_id

            if self.artifact_metadata:
                additional_args["Metadata"] = self.artifact_metadata

            print_progress_callback = ProgressPercentage(
                file_name, remote_path
            )
            future = self.transfer_manager.upload(
                file_name,
                self.bucket_name,
                remote_path,
                additional_args,
                [print_progress_callback],
            )
            future.result()

            return self.make_url(remote_path)

        except botocore.exceptions.ClientError as ex:
            error_code = ex.response["Error"]["Code"]
            if error_code == "NoSuchBucket":
                raise NoSuchBucketError(bucket_name=self.bucket_name)
            raise ex

    def upload_with_dedup(self, file_name, extension=None):
        """
        Makes and returns name of the S3 object based on the file's checksum

        :param file_name: file to upload
        :param extension: String of file extension to append to the object
        :return: S3 URL of the uploaded object
        """

        # This construction of remote_path is critical to preventing duplicate
        # uploads of same object. Uploader will check if the file exists in S3
        # and re-upload only if necessary. So the template points to same file
        # in multiple places, this will upload only once

        file_checksum = self.file_checksum(file_name)
        remote_path = file_checksum
        if extension:
            remote_path = remote_path + "." + extension

        return self.upload(file_name, remote_path)

    def file_exists(self, remote_path):
        """
        Check if the file we are trying to upload already exists in S3

        :param remote_path:
        :return: True, if file exists. False, otherwise
        """

        try:
            # Find the object that matches this ETag
            self.s3.head_object(Bucket=self.bucket_name, Key=remote_path)
            return True
        except botocore.exceptions.ClientError:
            # Either File does not exist or we are unable to get
            # this information.
            return False

    def make_url(self, obj_path):
        return f"s3://{self.bucket_name}/{obj_path}"

    def file_checksum(self, file_name):
        with open(file_name, "rb") as file_handle:
            checksum = _get_checksum()
            # Read file in chunks of 4096 bytes
            block_size = 4096

            # Save current cursor position and reset cursor to start of file
            curpos = file_handle.tell()
            file_handle.seek(0)

            buf = file_handle.read(block_size)
            while len(buf) > 0:
                checksum.update(buf)
                buf = file_handle.read(block_size)

            # Restore file cursor's position
            file_handle.seek(curpos)

            return checksum.hexdigest()

    def to_path_style_s3_url(self, key, version=None):
        """
        This link describes the format of Path Style URLs
        http://docs.aws.amazon.com/AmazonS3/latest/dev/UsingBucket.html#access-bucket-intro
        """
        base = self.s3.meta.endpoint_url
        result = f"{base}/{self.bucket_name}/{key}"
        if version:
            result = f"{result}?versionId={version}"

        return result


class ProgressPercentage(BaseSubscriber):
    # This class was copied directly from S3Transfer docs

    def __init__(self, filename, remote_path):
        self._filename = filename
        self._remote_path = remote_path
        self._size = float(os.path.getsize(filename))
        self._seen_so_far = 0
        self._lock = threading.Lock()

    def on_progress(self, future, bytes_transferred, **kwargs):
        # To simplify we'll assume this is hooked up
        # to a single filename.
        with self._lock:
            self._seen_so_far += bytes_transferred
            percentage = (self._seen_so_far / self._size) * 100
            sys.stderr.write(
                "\rUploading to %s  %s / %s  (%.2f%%)"
                % (
                    self._remote_path,
                    self._seen_so_far,
                    self._size,
                    percentage,
                )
            )
            sys.stderr.flush()



# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import io

from docutils.core import publish_string

from awscli.bcdoc import docevents, textwriter


class DocsGetter:
    """The main documentation getter for the auto-prompt workflow.

    This class calls out to helper classes to help grab the specific docs for
    service commands and service operations.

    """

    def __init__(self, driver):
        self._driver = driver
        self._cache = {}

    def _render_docs(self, help_command):
        renderer = FileRenderer()
        help_command.renderer = renderer
        help_command(None, None)
        # The report_level override is so that we don't print anything
        # to stdout/stderr on rendering issues.
        original_cli_help = renderer.contents.decode('utf-8')
        text_content = self._convert_rst_to_basic_text(original_cli_help)
        index = text_content.find('DESCRIPTION')
        if index > 0:
            text_content = text_content[index + len('DESCRIPTION') :]
        return text_content

    def _convert_rst_to_basic_text(self, contents):
        """Convert restructured text to basic text output.

        This function removes most of the decorations added
        in restructured text.

        This function is used to generate documentation we
        can show to users in a cross platform manner.

        Basic indentation and list formatting are kept,
        but many RST features are removed (such as
        section underlines).

        """
        # The report_level override is so that we don't print anything
        # to stdout/stderr on rendering issues.
        converted = publish_string(
            contents,
            writer=BasicTextWriter(),
            settings_overrides={'report_level': 5, 'halt_level': 5},
        )
        return converted.decode('utf-8').replace('\r', '')

    def get_doc_content(self, help_command):
        """Does the heavy lifting of retrieving the actual documentation
        content.

        """
        instance = help_command.EventHandlerClass(help_command)
        docevents.generate_events(help_command.session, help_command)
        content = self._render_docs(help_command)
        instance.unregister()
        return content

    def get_docs(self, parsed):
        lineage = parsed.lineage
        current_command = parsed.current_command
        if (tuple(lineage), current_command) not in self._cache:
            if current_command == 'aws':
                help_command = self._driver.create_help_command()
            else:
                subcommand_table = self._driver.subcommand_table
                for arg in lineage[1:]:
                    subcommand = subcommand_table[arg]
                    subcommand_table = subcommand.subcommand_table
                command = subcommand_table.get(current_command)
                help_command = command.create_help_command()
            docs = self.get_doc_content(help_command)
            self._cache[(tuple(lineage), current_command)] = docs
        return self._cache[(tuple(lineage), current_command)]


class FileRenderer:
    def __init__(self):
        self._io = io.BytesIO()

    def render(self, contents):
        self._io.write(contents)

    @property
    def contents(self):
        return self._io.getvalue()


class BasicTextWriter(textwriter.TextWriter):
    def translate(self):
        visitor = BasicTextTranslator(self.document)
        self.document.walkabout(visitor)
        self.output = visitor.body


class BasicTextTranslator(textwriter.TextTranslator):
    def depart_title(self, node):
        # Make the section titles upper cased, similar to
        # the man page output.
        text = ''.join(x[1] for x in self.states.pop() if x[0] == -1)
        self.stateindent.pop()
        self.states[-1].append((0, ['', text.upper(), '']))

    # The botocore TextWriter has additional formatting
    # for literals, for the aws-shell docs we don't want any
    # special processing so these nodes are noops.

    def visit_literal(self, node):
        pass

    def depart_literal(self, node):
        pass



# Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import logging

from prompt_toolkit.application import get_app
from prompt_toolkit.document import Document


class PromptToolkitHandler(logging.StreamHandler):
    def emit(self, record):
        try:
            app = get_app()
            if app.is_running and getattr(app, 'debug', False):
                msg = self.format(record)
                debug_buffer = app.layout.get_buffer_by_name('debug_buffer')
                current_document = debug_buffer.document.text
                if current_document:
                    msg = '\n'.join([current_document, msg])
                debug_buffer.set_document(
                    Document(text=msg), bypass_readonly=True
                )
            else:
                super().emit(record)
        except:
            self.handleError(record)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import logging
import os
import sys

from awscli.customizations.s3.syncstrategy.base import BaseSync
from awscli.customizations.utils import uni_print

LOG = logging.getLogger(__name__)


class CaseConflictException(Exception):
    pass


class CaseConflictSync(BaseSync):
    DOC_URI = (
        "https://docs.aws.amazon.com/cli/latest/topic/"
        "s3-case-insensitivity.html"
    )

    def __init__(
        self,
        sync_type='file_not_at_dest',
        on_case_conflict='ignore',
        submitted=None,
    ):
        super().__init__(sync_type)
        self._on_case_conflict = on_case_conflict
        if submitted is None:
            submitted = set()
        self._submitted = submitted

    @property
    def submitted(self):
        return self._submitted

    def determine_should_sync(self, src_file, dest_file):
        # `src_file.compare_key` and `dest_file.compare_key` are not equal.
        # This could mean that they're completely different or differ
        # only by case. eg, `/tmp/a` and `/tmp/b` versus `/tmp/a` and `/tmp/A`.
        # If the source file's destination already exists, that means it
        # differs only by case and the conflict needs to be handled.
        should_sync = True
        # Normalize compare key for case sensitivity.
        lower_compare_key = src_file.compare_key.lower()
        if lower_compare_key in self._submitted or os.path.exists(
            src_file.dest
        ):
            handler = getattr(self, f"_handle_{self._on_case_conflict}")
            should_sync = handler(src_file)
        if should_sync:
            LOG.debug(f"syncing: {src_file.src} -> {src_file.dest}")
            self._submitted.add(lower_compare_key)
            # Set properties so that a subscriber can be created
            # that removes the key from the set after download finishes.
            src_file.case_conflict_submitted = self._submitted
            src_file.case_conflict_key = lower_compare_key
        return should_sync

    @staticmethod
    def _handle_ignore(src_file):
        return True

    @staticmethod
    def _handle_skip(src_file):
        msg = (
            f"warning: Skipping {src_file.src} -> {src_file.dest} "
            "because a file whose name differs only by case either exists "
            "or is being downloaded.\n"
        )
        uni_print(msg, sys.stderr)
        return False

    @staticmethod
    def _handle_warn(src_file):
        msg = (
            f"warning: Downloading {src_file.src} -> {src_file.dest} "
            "despite a file whose name differs only by case either existing "
            "or being downloaded. This behavior is not defined on "
            "case-insensitive filesystems and may result in overwriting "
            "existing files or race conditions between concurrent downloads. "
            f"For more information, see {CaseConflictSync.DOC_URI}.\n"
        )
        uni_print(msg, sys.stderr)
        return True

    @staticmethod
    def _handle_error(src_file):
        msg = (
            f"Failed to download {src_file.src} -> {src_file.dest} "
            "because a file whose name differs only by case either exists "
            "or is being downloaded."
        )
        raise CaseConflictException(msg)



# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import copy
import math

from botocore.exceptions import ClientError
from s3transfer.exceptions import S3CopyFailedError
from s3transfer.tasks import (
    CompleteMultipartUploadTask,
    CreateMultipartUploadTask,
    SubmissionTask,
    Task,
)
from s3transfer.utils import (
    ChunksizeAdjuster,
    calculate_range_parameter,
    get_callbacks,
    get_filtered_dict,
)


class CopySubmissionTask(SubmissionTask):
    """Task for submitting tasks to execute a copy"""

    EXTRA_ARGS_TO_HEAD_ARGS_MAPPING = {
        'CopySourceIfMatch': 'IfMatch',
        'CopySourceIfModifiedSince': 'IfModifiedSince',
        'CopySourceIfNoneMatch': 'IfNoneMatch',
        'CopySourceIfUnmodifiedSince': 'IfUnmodifiedSince',
        'CopySourceSSECustomerKey': 'SSECustomerKey',
        'CopySourceSSECustomerAlgorithm': 'SSECustomerAlgorithm',
        'CopySourceSSECustomerKeyMD5': 'SSECustomerKeyMD5',
        'RequestPayer': 'RequestPayer',
        'ExpectedBucketOwner': 'ExpectedBucketOwner',
    }

    UPLOAD_PART_COPY_ARGS = [
        'CopySourceIfMatch',
        'CopySourceIfModifiedSince',
        'CopySourceIfNoneMatch',
        'CopySourceIfUnmodifiedSince',
        'CopySourceSSECustomerKey',
        'CopySourceSSECustomerAlgorithm',
        'CopySourceSSECustomerKeyMD5',
        'SSECustomerKey',
        'SSECustomerAlgorithm',
        'SSECustomerKeyMD5',
        'RequestPayer',
        'ExpectedBucketOwner',
    ]

    CREATE_MULTIPART_ARGS_BLACKLIST = [
        'CopySourceIfMatch',
        'CopySourceIfModifiedSince',
        'CopySourceIfNoneMatch',
        'CopySourceIfUnmodifiedSince',
        'CopySourceSSECustomerKey',
        'CopySourceSSECustomerAlgorithm',
        'CopySourceSSECustomerKeyMD5',
        'MetadataDirective',
        'TaggingDirective',
        'IfNoneMatch',
    ]

    COMPLETE_MULTIPART_ARGS = [
        'SSECustomerKey',
        'SSECustomerAlgorithm',
        'SSECustomerKeyMD5',
        'RequestPayer',
        'ExpectedBucketOwner',
        'IfNoneMatch',
    ]


    def _submit(
        self, client, config, osutil, request_executor, transfer_future
    ):
        """
        :param client: The client associated with the transfer manager

        :type config: s3transfer.manager.TransferConfig
        :param config: The transfer config associated with the transfer
            manager

        :type osutil: s3transfer.utils.OSUtil
        :param osutil: The os utility associated to the transfer manager

        :type request_executor: s3transfer.futures.BoundedExecutor
        :param request_executor: The request executor associated with the
            transfer manager

        :type transfer_future: s3transfer.futures.TransferFuture
        :param transfer_future: The transfer future associated with the
            transfer request that tasks are being submitted for
        """
        if (
            transfer_future.meta.size is None
            or transfer_future.meta.etag is None
        ):
            # If a size was not provided figure out the size for the
            # user. Note that we will only use the client provided to
            # the TransferManager. If the object is outside of the region
            # of the client, they may have to provide the file size themselves
            # with a completely new client.
            call_args = transfer_future.meta.call_args
            head_object_request = (
                self._get_head_object_request_from_copy_source(
                    call_args.copy_source
                )
            )
            extra_args = call_args.extra_args

            # Map any values that may be used in the head object that is
            # used in the copy object
            for param, value in extra_args.items():
                if param in self.EXTRA_ARGS_TO_HEAD_ARGS_MAPPING:
                    head_object_request[
                        self.EXTRA_ARGS_TO_HEAD_ARGS_MAPPING[param]
                    ] = value

            response = call_args.source_client.head_object(
                **head_object_request
            )
            transfer_future.meta.provide_transfer_size(
                response['ContentLength']
            )
            # Provide an etag to ensure a stored object is not modified
            # during a multipart copy.
            transfer_future.meta.provide_object_etag(response.get('ETag'))

        # If it is greater than threshold do a multipart copy, otherwise
        # do a regular copy object.
        if transfer_future.meta.size < config.multipart_threshold:
            self._submit_copy_request(
                client, config, osutil, request_executor, transfer_future
            )
        else:
            self._submit_multipart_request(
                client, config, osutil, request_executor, transfer_future
            )

    def _submit_copy_request(
        self, client, config, osutil, request_executor, transfer_future
    ):
        call_args = transfer_future.meta.call_args

        # Get the needed progress callbacks for the task
        progress_callbacks = get_callbacks(transfer_future, 'progress')

        self._transfer_coordinator.submit(
            request_executor,
            CopyObjectTask(
                transfer_coordinator=self._transfer_coordinator,
                main_kwargs={
                    "client": client,
                    "copy_source": call_args.copy_source,
                    "bucket": call_args.bucket,
                    "key": call_args.key,
                    "extra_args": call_args.extra_args,
                    "callbacks": progress_callbacks,
                    "size": transfer_future.meta.size,
                },
                is_final=True,
            ),
        )

    def _submit_multipart_request(
        self, client, config, osutil, request_executor, transfer_future
    ):
        call_args = transfer_future.meta.call_args

        # Submit the request to create a multipart upload and make sure it
        # does not include any of the arguments used for copy part.
        create_multipart_extra_args = {}
        for param, val in call_args.extra_args.items():
            if param not in self.CREATE_MULTIPART_ARGS_BLACKLIST:
                create_multipart_extra_args[param] = val

        create_multipart_future = self._transfer_coordinator.submit(
            request_executor,
            CreateMultipartUploadTask(
                transfer_coordinator=self._transfer_coordinator,
                main_kwargs={
                    'client': client,
                    'bucket': call_args.bucket,
                    'key': call_args.key,
                    'extra_args': create_multipart_extra_args,
                },
            ),
        )

        # Determine how many parts are needed based on filesize and
        # desired chunksize.
        part_size = config.multipart_chunksize
        adjuster = ChunksizeAdjuster()
        part_size = adjuster.adjust_chunksize(
            part_size, transfer_future.meta.size
        )
        num_parts = int(
            math.ceil(transfer_future.meta.size / float(part_size))
        )

        # Submit requests to upload the parts of the file.
        part_futures = []
        progress_callbacks = get_callbacks(transfer_future, 'progress')

        for part_number in range(1, num_parts + 1):
            extra_part_args = self._extra_upload_part_args(
                call_args.extra_args
            )
            # The part number for upload part starts at 1 while the
            # range parameter starts at zero, so just subtract 1 off of
            # the part number
            extra_part_args['CopySourceRange'] = calculate_range_parameter(
                part_size,
                part_number - 1,
                num_parts,
                transfer_future.meta.size,
            )
            if transfer_future.meta.etag is not None:
                extra_part_args['CopySourceIfMatch'] = (
                    transfer_future.meta.etag
                )
            # Get the size of the part copy as well for the progress
            # callbacks.
            size = self._get_transfer_size(
                part_size,
                part_number - 1,
                num_parts,
                transfer_future.meta.size,
            )
            # Get the checksum algorithm of the multipart request.
            checksum_algorithm = call_args.extra_args.get("ChecksumAlgorithm")
            part_futures.append(
                self._transfer_coordinator.submit(
                    request_executor,
                    CopyPartTask(
                        transfer_coordinator=self._transfer_coordinator,
                        main_kwargs={
                            'client': client,
                            'copy_source': call_args.copy_source,
                            'bucket': call_args.bucket,
                            'key': call_args.key,
                            'part_number': part_number,
                            'extra_args': extra_part_args,
                            'callbacks': progress_callbacks,
                            'size': size,
                            'checksum_algorithm': checksum_algorithm,
                        },
                        pending_main_kwargs={
                            'upload_id': create_multipart_future
                        },
                    ),
                )
            )

        complete_multipart_extra_args = self._extra_complete_multipart_args(
            call_args.extra_args
        )
        # Submit the request to complete the multipart upload.
        self._transfer_coordinator.submit(
            request_executor,
            CompleteMultipartUploadTask(
                transfer_coordinator=self._transfer_coordinator,
                main_kwargs={
                    'client': client,
                    'bucket': call_args.bucket,
                    'key': call_args.key,
                    'extra_args': complete_multipart_extra_args,
                },
                pending_main_kwargs={
                    'upload_id': create_multipart_future,
                    'parts': part_futures,
                },
                is_final=True,
            ),
        )

    def _get_head_object_request_from_copy_source(self, copy_source):
        if isinstance(copy_source, dict):
            return copy.copy(copy_source)
        else:
            raise TypeError(
                'Expecting dictionary formatted: '
                '{"Bucket": bucket_name, "Key": key} '
                f'but got {copy_source} or type {type(copy_source)}.'
            )

    def _extra_upload_part_args(self, extra_args):
        # Only the args in COPY_PART_ARGS actually need to be passed
        # onto the upload_part_copy calls.
        return get_filtered_dict(extra_args, self.UPLOAD_PART_COPY_ARGS)

    def _extra_complete_multipart_args(self, extra_args):
        return get_filtered_dict(extra_args, self.COMPLETE_MULTIPART_ARGS)

    def _get_transfer_size(
        self, part_size, part_index, num_parts, total_transfer_size
    ):
        if part_index == num_parts - 1:
            # The last part may be different in size then the rest of the
            # parts.
            return total_transfer_size - (part_index * part_size)
        return part_size


class CopyObjectTask(Task):
    """Task to do a nonmultipart copy"""

    def _main(
        self, client, copy_source, bucket, key, extra_args, callbacks, size
    ):
        """
        :param client: The client to use when calling PutObject
        :param copy_source: The CopySource parameter to use
        :param bucket: The name of the bucket to copy to
        :param key: The name of the key to copy to
        :param extra_args: A dictionary of any extra arguments that may be
            used in the upload.
        :param callbacks: List of callbacks to call after copy
        :param size: The size of the transfer. This value is passed into
            the callbacks

        """
        client.copy_object(
            CopySource=copy_source, Bucket=bucket, Key=key, **extra_args
        )
        for callback in callbacks:
            callback(bytes_transferred=size)


class CopyPartTask(Task):
    """Task to upload a part in a multipart copy"""

    def _main(
        self,
        client,
        copy_source,
        bucket,
        key,
        upload_id,
        part_number,
        extra_args,
        callbacks,
        size,
        checksum_algorithm=None,
    ):
        """
        :param client: The client to use when calling PutObject
        :param copy_source: The CopySource parameter to use
        :param bucket: The name of the bucket to upload to
        :param key: The name of the key to upload to
        :param upload_id: The id of the upload
        :param part_number: The number representing the part of the multipart
            upload
        :param extra_args: A dictionary of any extra arguments that may be
            used in the upload.
        :param callbacks: List of callbacks to call after copy part
        :param size: The size of the transfer. This value is passed into
            the callbacks
        :param checksum_algorithm: The algorithm that was used to create the multipart
            upload

        :rtype: dict
        :returns: A dictionary representing a part::

            {'Etag': etag_value, 'PartNumber': part_number}

            This value can be appended to a list to be used to complete
            the multipart upload. If a checksum is in the response,
            it will also be included.
        """
        try:
            response = client.upload_part_copy(
                CopySource=copy_source,
                Bucket=bucket,
                Key=key,
                UploadId=upload_id,
                PartNumber=part_number,
                **extra_args,
            )
        except ClientError as e:
            error_code = e.response.get('Error', {}).get('Code')
            src_key = copy_source['Key']
            src_bucket = copy_source['Bucket']
            if error_code == "PreconditionFailed":
                raise S3CopyFailedError(
                    f'Contents of stored object "{src_key}" '
                    f'in bucket "{src_bucket}" did not match '
                    'expected ETag.'
                )
            else:
                raise
        for callback in callbacks:
            callback(bytes_transferred=size)
        etag = response['CopyPartResult']['ETag']
        part_metadata = {'ETag': etag, 'PartNumber': part_number}
        if checksum_algorithm:
            checksum_member = f'Checksum{checksum_algorithm.upper()}'
            if checksum_member in response['CopyPartResult']:
                part_metadata[checksum_member] = response['CopyPartResult'][
                    checksum_member
                ]
        return part_metadata



# Copyright 2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from awscli.customizations.exceptions import ParamValidationError


class EmrError(Exception):
    """
    The base exception class for Emr exceptions.

    :ivar msg: The descriptive message associated with the error.
    """

    fmt = 'An unspecified error occurred'

    def __init__(self, **kwargs):
        msg = self.fmt.format(**kwargs)
        Exception.__init__(self, msg)
        self.kwargs = kwargs


class MissingParametersError(EmrError, ParamValidationError):
    """
    One or more required parameters were not supplied.

    :ivar object_name: The object that has missing parameters.
        This can be an operation or a parameter (in the
        case of inner params).  The str() of this object
        will be used so it doesn't need to implement anything
        other than str().
    :ivar missing: The names of the missing parameters.
    """

    fmt = (
        'The following required parameters are missing for '
        '{object_name}: {missing}.'
    )


class EmptyListError(EmrError, ParamValidationError):
    """
    The provided list is empty.

    :ivar param: The provided list parameter
    """

    fmt = 'The parameter {param} cannot be an empty list.'


class MissingRequiredInstanceGroupsError(EmrError, ParamValidationError):
    """
    In create-cluster command, none of --instance-group,
    --instance-count nor --instance-type were not supplied.
    """

    fmt = (
        'Must specify either --instance-groups or '
        '--instance-type with --instance-count(optional) to '
        'configure instance groups.'
    )


class InstanceGroupsValidationError(EmrError, ParamValidationError):
    """
    --instance-type and --instance-count are shortcut option
    for --instance-groups and they cannot be specified
    together with --instance-groups
    """

    fmt = (
        'You may not specify --instance-type '
        'or --instance-count with --instance-groups, '
        'because --instance-type and --instance-count are '
        'shortcut options for --instance-groups.'
    )


class InvalidAmiVersionError(EmrError, ParamValidationError):
    """
    The supplied ami-version is invalid.
    :ivar ami_version: The provided ami_version.
    """

    fmt = (
        'The supplied AMI version "{ami_version}" is invalid.'
        ' Please see AMI Versions Supported in Amazon EMR in '
        'Amazon Elastic MapReduce Developer Guide: '
        'http://docs.aws.amazon.com/ElasticMapReduce/'
        'latest/DeveloperGuide/ami-versions-supported.html'
    )


class MissingBooleanOptionsError(EmrError, ParamValidationError):
    """
    Required boolean options are not supplied.

    :ivar true_option
    :ivar false_option
    """

    fmt = (
        'Must specify one of the following boolean options: '
        '{true_option}|{false_option}.'
    )


class UnknownStepTypeError(EmrError, ParamValidationError):
    """
    The provided step type is not supported.

    :ivar step_type: the step_type provided.
    """

    fmt = 'The step type {step_type} is not supported.'


class UnknownIamEndpointError(EmrError):
    """
    The IAM endpoint is not known for the specified region.

    :ivar region: The region specified.
    """

    fmt = (
        'IAM endpoint not known for region: {region}.'
        + ' Specify the iam-endpoint using the --iam-endpoint option.'
    )


class ResolveServicePrincipalError(EmrError):
    """
    The service principal could not be resolved from the region or the
    endpoint.
    """

    fmt = (
        'Could not resolve the service principal from'
        + ' the region or the endpoint.'
    )


class LogUriError(EmrError, ParamValidationError):
    """
    The LogUri is not specified and debugging is enabled for the cluster.
    """

    fmt = (
        'LogUri not specified. You must specify a logUri '
        'if you enable debugging when creating a cluster.'
    )


class MasterDNSNotAvailableError(EmrError):
    """
    Cannot get dns of master node on the cluster.
    """

    fmt = (
        'Cannot get DNS of master node on the cluster. '
        ' Please try again after some time.'
    )


class WrongPuttyKeyError(EmrError, ParamValidationError):
    """
    A wrong key has been used with a compatible program.
    """

    fmt = (
        'Key file file format is incorrect. Putty expects a ppk file. '
        'Please refer to documentation at http://docs.aws.amazon.com/'
        'ElasticMapReduce/latest/DeveloperGuide/EMR_SetUp_SSH.html. '
    )


class SSHNotFoundError(EmrError):
    """
    SSH or Putty not available.
    """

    fmt = (
        'SSH or Putty not available. Please refer to the documentation '
        'at http://docs.aws.amazon.com/ElasticMapReduce/latest/'
        'DeveloperGuide/EMR_SetUp_SSH.html.'
    )


class SCPNotFoundError(EmrError):
    """
    SCP or Pscp not available.
    """

    fmt = (
        'SCP or Pscp not available. Please refer to the documentation '
        'at http://docs.aws.amazon.com/ElasticMapReduce/latest/'
        'DeveloperGuide/EMR_SetUp_SSH.html. '
    )


class SubnetAndAzValidationError(EmrError, ParamValidationError):
    """
    SubnetId and AvailabilityZone are mutual exclusive in --ec2-attributes.
    """

    fmt = (
        'You may not specify both a SubnetId and an Availabili'
        'tyZone (placement) because ec2SubnetId implies a placement.'
    )


class RequiredOptionsError(EmrError, ParamValidationError):
    """
    Either of option1 or option2 is required.
    """

    fmt = 'Either {option1} or {option2} is required.'


class MutualExclusiveOptionError(EmrError, ParamValidationError):
    """
    The provided option1 and option2 are mutually exclusive.

    :ivar option1
    :ivar option2
    :ivar message (optional)
    """

    def __init__(self, **kwargs):
        msg = (
            'You cannot specify both '
            + kwargs.get('option1', '')
            + ' and '
            + kwargs.get('option2', '')
            + ' options together.'
            + kwargs.get('message', '')
        )
        Exception.__init__(self, msg)


class MissingApplicationsError(EmrError, ParamValidationError):
    """
    The application required for a step is not installed when creating a
    cluster.

    :ivar applications
    """

    def __init__(self, **kwargs):
        msg = (
            'Some of the steps require the following'
            ' applications to be installed: '
            + ', '.join(kwargs['applications'])
            + '. Please install the'
            ' applications using --applications.'
        )
        Exception.__init__(self, msg)


class ClusterTerminatedError(EmrError):
    """
    The cluster is terminating or has already terminated.
    """

    fmt = 'Cluster terminating or already terminated.'


class ClusterStatesFilterValidationError(EmrError, ParamValidationError):
    """
    In the list-clusters command, customers can specify only one
    of the following states filters:
    --cluster-states, --active, --terminated, --failed

    """

    fmt = (
        'You can specify only one of the cluster state '
        'filters: --cluster-states, --active, --terminated, --failed.'
    )


class MissingClusterAttributesError(EmrError, ParamValidationError):
    """
    In the modify-cluster-attributes command, customers need to provide
    at least one of the following cluster attributes: --visible-to-all-users,
    --no-visible-to-all-users, --termination-protected, --no-termination-protected,
    --auto-terminate and --no-auto-terminate
    """

    fmt = (
        'Must specify one of the following boolean options: '
        '--visible-to-all-users|--no-visible-to-all-users, '
        '--termination-protected|--no-termination-protected, '
        '--auto-terminate|--no-auto-terminate, '
        '--unhealthy-node-replacement|--no-unhealthy-node-replacement.'
    )


class InvalidEmrFsArgumentsError(EmrError, ParamValidationError):
    """
    The provided EMRFS parameters are invalid as parent feature e.g.,
    Consistent View, CSE, SSE is not configured

    :ivar invalid: Invalid parameters
    :ivar parent_object_name: Parent feature name
    """

    fmt = (
        '{parent_object_name} is not specified. Thus, '
        ' following parameters are invalid: {invalid}'
    )


class DuplicateEmrFsConfigurationError(EmrError, ParamValidationError):
    fmt = (
        'EMRFS should be configured either using '
        '--configuration or --emrfs but not both'
    )


class UnknownCseProviderTypeError(EmrError, ParamValidationError):
    """
    The provided EMRFS client-side encryption provider type is not supported.

    :ivar provider_type: the provider_type provided.
    """

    fmt = (
        'The client side encryption type "{provider_type}" is '
        'not supported. You must specify either KMS or Custom'
    )


class UnknownEncryptionTypeError(EmrError, ParamValidationError):
    """
    The provided encryption type is not supported.

    :ivar provider_type: the provider_type provided.
    """

    fmt = (
        'The encryption type "{encryption}" is invalid. '
        'You must specify either ServerSide or ClientSide'
    )


class BothSseAndEncryptionConfiguredError(EmrError, ParamValidationError):
    """
    Only one of SSE or Encryption can be configured.

    :ivar sse: Value for SSE
    :ivar encryption: Value for encryption
    """

    fmt = (
        'Both SSE={sse} and Encryption={encryption} are '
        'configured for --emrfs. You must specify only one of the two.'
    )


class InvalidBooleanConfigError(EmrError, ParamValidationError):
    fmt = (
        "{config_value} for {config_key} in the config file is "
        "invalid. The value should be either 'True' or 'False'. Use "
        "'aws configure set {profile_var_name}.emr.{config_key} <value>' "
        "command to set a valid value."
    )


class UnsupportedCommandWithReleaseError(EmrError, ParamValidationError):
    fmt = (
        "{command} is not supported with "
        "'{release_label}' release."
    )


class MissingAutoScalingRoleError(EmrError, ParamValidationError):
    fmt = (
        "Must specify --auto-scaling-role when configuring an "
        "AutoScaling policy for an instance group."
    )

class InvalidS3LoggingLogTypeError(EmrError, ParamValidationError):
    """
    Invalid log type specified for S3LoggingConfiguration.
    """
    fmt = ('Invalid log type specified for the current '
           'S3LoggingConfiguration. Supported log types: system-logs, '
           'application-logs, persistent-ui-logs')

class InvalidS3LoggingPolicyError(EmrError, ParamValidationError):
    """
    Invalid policy specified for S3LoggingConfiguration.
    """
    fmt = ('Invalid policy specified for the current '
           'S3LoggingConfiguration. Supported policies: emr-managed, '
           'on-customer-s3only, disabled')

class InvalidS3LoggingPersistentUiLogsPolicyError(EmrError, ParamValidationError):
    """
    Invalid policy for persistent-ui-logs in S3LoggingConfiguration.
    """
    fmt = ("Invalid policy for log type 'persistent-ui-logs'. "
           "Supported values for persistent-ui-logs: emr-managed, disabled")

class S3LoggingConfigurationLogUriRequiredError(EmrError, ParamValidationError):
    """
    LogUri is required for the current S3LoggingConfiguration.
    """
    fmt = ('A valid S3 location (LogUri) is required for the '
           'current S3LoggingConfiguration. Please specify an S3 bucket and '
           'try again.')

class S3LoggingConfigurationLogUriNotAllowedError(EmrError, ParamValidationError):
    """
    LogUri must not be specified when system-logs and application-logs
    policies are both disabled in S3LoggingConfiguration.
    """
    fmt = ('LogUri must not be specified when system-logs and '
           'application-logs policies are both disabled in '
           'S3LoggingConfiguration')


# Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.


class EKS:
    def __init__(self, eks_client):
        self.eks_client = eks_client
        self.cluster_info = {}

    def get_oidc_issuer_id(self, cluster_name):
        """Method to get OIDC issuer id for the given EKS cluster"""
        if cluster_name not in self.cluster_info:
            self.cluster_info[cluster_name] = self.eks_client.describe_cluster(
                name=cluster_name
            )

        oidc_issuer = (
            self.cluster_info[cluster_name]
            .get("cluster", {})
            .get("identity", {})
            .get("oidc", {})
            .get("issuer", "")
        )

        return oidc_issuer.split('https://')[1]

    def get_account_id(self, cluster_name):
        """Method to get account id for the given EKS cluster"""
        if cluster_name not in self.cluster_info:
            self.cluster_info[cluster_name] = self.eks_client.describe_cluster(
                name=cluster_name
            )

        cluster_arn = (
            self.cluster_info[cluster_name].get("cluster", {}).get("arn", "")
        )

        return cluster_arn.split(':')[4]

    def create_pod_identity_association(
        self, cluster_name, namespace, role_arn, service_account
    ):
        return self.eks_client.create_pod_identity_association(
            clusterName=cluster_name,
            namespace=namespace,
            roleArn=role_arn,
            serviceAccount=service_account,
        )

    def delete_pod_identity_association(self, cluster_name, association_id):
        return self.eks_client.delete_pod_identity_association(
            clusterName=cluster_name, associationId=association_id
        )

    def list_pod_identity_associations(
        self, cluster_name, namespace, service_account
    ):
        return self.eks_client.list_pod_identity_associations(
            clusterName=cluster_name,
            namespace=namespace,
            serviceAccount=service_account,
        )



# Copyright 2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import fnmatch
import logging
import os

from awscli.customizations.s3.utils import split_s3_bucket_key

LOG = logging.getLogger(__name__)


def create_filter(parameters):
    """Given the CLI parameters dict, create a Filter object."""
    # We need to evaluate all the filters based on the source
    # directory.
    if parameters['filters']:
        cli_filters = parameters['filters']
        real_filters = []
        for filter_type, filter_pattern in cli_filters:
            real_filters.append((filter_type.lstrip('-'), filter_pattern))
        source_location = parameters['src']
        if source_location.startswith('s3://'):
            # This gives us (bucket, keyname) and we want
            # the bucket to be the root dir.
            src_rootdir = _get_s3_root(source_location, parameters['dir_op'])
        else:
            src_rootdir = _get_local_root(
                parameters['src'], parameters['dir_op']
            )

        destination_location = parameters['dest']
        if destination_location.startswith('s3://'):
            dst_rootdir = _get_s3_root(
                parameters['dest'], parameters['dir_op']
            )
        else:
            dst_rootdir = _get_local_root(
                parameters['dest'], parameters['dir_op']
            )

        return Filter(real_filters, src_rootdir, dst_rootdir)
    else:
        return Filter({}, None, None)


def _get_s3_root(source_location, dir_op):
    # Obtain the bucket and the key.
    bucket, key = split_s3_bucket_key(source_location)
    if not dir_op and not key.endswith('/'):
        # If we are not performing an operation on a directory and the key
        # is of the form: ``prefix/key``. We only want ``prefix`` included in
        # the the s3 root and not ``key``.
        key = '/'.join(key.split('/')[:-1])
    # Rejoin the bucket and key back together.
    s3_path = '/'.join([bucket, key])
    return s3_path


def _get_local_root(source_location, dir_op):
    if dir_op:
        rootdir = os.path.abspath(source_location)
    else:
        rootdir = os.path.abspath(os.path.dirname(source_location))
    return rootdir


class Filter:
    """
    This is a universal exclude/include filter.
    """

    def __init__(self, patterns, rootdir, dst_rootdir):
        """
        :var patterns: A list of patterns. A pattern consists of a list
            whose first member is a string 'exclude' or 'include'.
            The second member is the actual rule.
        :var rootdir: The root directory where the patterns are evaluated.
            This will generally be the directory of the source location.
        :var dst_rootdir: The destination root directory where the patterns are
            evaluated.  This is only useful when the --delete option is
            also specified.

        """
        self._original_patterns = patterns
        self.patterns = self._full_path_patterns(patterns, rootdir)
        self.dst_patterns = self._full_path_patterns(patterns, dst_rootdir)

    def _full_path_patterns(self, original_patterns, rootdir):
        # We need to transform the patterns into patterns that have
        # the root dir prefixed, so things like ``--exclude "*"``
        # will actually be ['exclude', '/path/to/root/*']
        full_patterns = []
        for pattern in original_patterns:
            full_patterns.append(
                (pattern[0], os.path.join(rootdir, pattern[1]))
            )
        return full_patterns

    def call(self, file_infos):
        """
        This function iterates over through the yielded file_info objects.  It
        determines the type of the file and applies pattern matching to
        determine if the rule applies.  While iterating though the patterns the
        file is assigned a boolean flag to determine if a file should be
        yielded on past the filer.  Anything identified by the exclude filter
        has its flag set to false.  Anything identified by the include filter
        has its flag set to True.  All files begin with the flag set to true.
        Rules listed at the end will overwrite flags thrown by rules listed
        before it.
        """
        for file_info in file_infos:
            file_path = file_info.src
            file_status = (file_info, True)
            for pattern, dst_pattern in zip(self.patterns, self.dst_patterns):
                current_file_status = self._match_pattern(pattern, file_info)
                if current_file_status is not None:
                    file_status = current_file_status
                dst_current_file_status = self._match_pattern(
                    dst_pattern, file_info
                )
                if dst_current_file_status is not None:
                    file_status = dst_current_file_status
            LOG.debug(
                "=%s final filtered status, should_include: %s",
                file_path,
                file_status[1],
            )
            if file_status[1]:
                yield file_info

    def _match_pattern(self, pattern, file_info):
        file_status = None
        file_path = file_info.src
        pattern_type = pattern[0]
        if file_info.src_type == 'local':
            path_pattern = pattern[1].replace('/', os.sep)
        else:
            path_pattern = pattern[1].replace(os.sep, '/')
        is_match = fnmatch.fnmatch(file_path, path_pattern)
        if is_match and pattern_type == 'include':
            file_status = (file_info, True)
            LOG.debug("%s matched include filter: %s", file_path, path_pattern)
        elif is_match and pattern_type == 'exclude':
            file_status = (file_info, False)
            LOG.debug("%s matched exclude filter: %s", file_path, path_pattern)
        else:
            LOG.debug(
                "%s did not match %s filter: %s",
                file_path,
                pattern_type,
                path_pattern,
            )
        return file_status



# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from awscli.customizations.sso.utils import (
    LOGIN_ARGS,
    BaseSSOCommand,
    PrintOnlyHandler,
    do_sso_login,
)
from awscli.customizations.utils import uni_print


class LoginCommand(BaseSSOCommand):
    NAME = 'login'
    DESCRIPTION = (
        'Retrieves and caches an AWS IAM Identity Center access token to '
        'exchange for AWS credentials. To login, the requested profile must '
        'have first been setup using ``aws configure sso``. Each time the '
        '``login`` command is called, a new SSO access token will be '
        'retrieved. Please note that only one login session can be active for '
        'a given SSO Session and creating multiple profiles does not allow for'
        ' multiple users to be authenticated against the same SSO Session.'
    )
    ARG_TABLE = LOGIN_ARGS + [
        {
            'name': 'sso-session',
            'help_text': (
                'An explicit SSO session to use to login. By default, this '
                'command will login using the SSO session configured as part '
                'of the requested profile and generally does not require this '
                'argument to be set.'
            ),
        }
    ]

    def _run_main(self, parsed_args, parsed_globals):
        sso_config = self._get_sso_config(sso_session=parsed_args.sso_session)
        on_pending_authorization = None
        if parsed_args.no_browser:
            on_pending_authorization = PrintOnlyHandler()
        do_sso_login(
            session=self._session,
            parsed_globals=parsed_globals,
            sso_region=sso_config['sso_region'],
            start_url=sso_config['sso_start_url'],
            on_pending_authorization=on_pending_authorization,
            force_refresh=True,
            session_name=sso_config.get('session_name'),
            registration_scopes=sso_config.get('registration_scopes'),
            use_device_code=parsed_args.use_device_code,
        )
        success_msg = 'Successfully logged into Start URL: %s\n'
        uni_print(success_msg % sso_config['sso_start_url'])
        return 0



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import errno
import json
import logging
import os
import re
from subprocess import check_call, check_output

from awscli.clidriver import CLIOperationCaller, ServiceOperation
from awscli.compat import ignore_user_entered_signals

logger = logging.getLogger(__name__)

ERROR_MESSAGE = (
    'SessionManagerPlugin is not found. ',
    'Please refer to SessionManager Documentation here: ',
    'http://docs.aws.amazon.com/console/systems-manager/',
    'session-manager-plugin-not-found',
)


def register_ssm_session(event_handlers):
    event_handlers.register(
        'building-command-table.ssm', add_custom_start_session
    )


def add_custom_start_session(session, command_table, **kwargs):
    command_table['start-session'] = StartSessionCommand(
        name='start-session',
        parent_name='ssm',
        session=session,
        operation_model=session.get_service_model('ssm').operation_model(
            'StartSession'
        ),
        operation_caller=StartSessionCaller(session),
    )


class VersionRequirement:
    WHITESPACE_REGEX = re.compile(r"\s+")
    SSM_SESSION_PLUGIN_VERSION_REGEX = re.compile(r"^\d+(\.\d+){0,3}$")

    def __init__(self, min_version):
        self.min_version = min_version

    def meets_requirement(self, version):
        ssm_plugin_version = self._sanitize_plugin_version(version)
        if self._is_valid_version(ssm_plugin_version):
            norm_version, norm_min_version = self._normalize(
                ssm_plugin_version, self.min_version
            )
            return norm_version > norm_min_version
        else:
            return False

    def _sanitize_plugin_version(self, plugin_version):
        return re.sub(self.WHITESPACE_REGEX, "", plugin_version)

    def _is_valid_version(self, plugin_version):
        return bool(
            self.SSM_SESSION_PLUGIN_VERSION_REGEX.match(plugin_version)
        )

    def _normalize(self, v1, v2):
        v1_parts = [int(v) for v in v1.split(".")]
        v2_parts = [int(v) for v in v2.split(".")]
        while len(v1_parts) != len(v2_parts):
            if len(v1_parts) - len(v2_parts) > 0:
                v2_parts.append(0)
            else:
                v1_parts.append(0)
        return v1_parts, v2_parts


class StartSessionCommand(ServiceOperation):
    def create_help_command(self):
        help_command = super(StartSessionCommand, self).create_help_command()
        # Change the output shape because the command provides no output.
        self._operation_model.output_shape = None
        return help_command


class StartSessionCaller(CLIOperationCaller):
    LAST_PLUGIN_VERSION_WITHOUT_ENV_VAR = "1.2.497.0"
    DEFAULT_SSM_ENV_NAME = "AWS_SSM_START_SESSION_RESPONSE"

    def invoke(self, service_name, operation_name, parameters, parsed_globals):
        client = self._session.create_client(
            service_name,
            region_name=parsed_globals.region,
            endpoint_url=parsed_globals.endpoint_url,
            verify=parsed_globals.verify_ssl,
        )
        response = client.start_session(**parameters)
        session_id = response['SessionId']
        region_name = client.meta.region_name
        # Profile_name is used to passed on to session manager plugin
        # to fetch same profile credentials to make an api call in the plugin.
        # If --profile flag is configured, pass it to Session Manager plugin.
        # If not, set empty string.
        profile_name = (
            parsed_globals.profile
            if parsed_globals.profile is not None
            else ''
        )
        endpoint_url = client.meta.endpoint_url
        ssm_env_name = self.DEFAULT_SSM_ENV_NAME

        try:
            session_parameters = {
                "SessionId": response["SessionId"],
                "TokenValue": response["TokenValue"],
                "StreamUrl": response["StreamUrl"],
            }
            start_session_response = json.dumps(session_parameters)

            plugin_version = check_output(
                ["session-manager-plugin", "--version"], text=True
            )
            env = os.environ.copy()

            # Check if this plugin supports passing the start session response
            # as an environment variable name. If it does, it will set the
            # value to the response from the start_session operation to the env
            # variable defined in DEFAULT_SSM_ENV_NAME. If the session plugin
            # version is invalid or older than the version defined in
            # LAST_PLUGIN_VERSION_WITHOUT_ENV_VAR, it will fall back to
            # passing the start_session response directly.
            version_requirement = VersionRequirement(
                min_version=self.LAST_PLUGIN_VERSION_WITHOUT_ENV_VAR
            )
            if version_requirement.meets_requirement(plugin_version):
                env[ssm_env_name] = start_session_response
                start_session_response = ssm_env_name
            # ignore_user_entered_signals ignores these signals
            # because if signals which kills the process are not
            # captured would kill the foreground process but not the
            # background one. Capturing these would prevents process
            # from getting killed and these signals are input to plugin
            # and handling in there
            with ignore_user_entered_signals():
                # call executable with necessary input
                check_call(
                    [
                        "session-manager-plugin",
                        start_session_response,
                        region_name,
                        "StartSession",
                        profile_name,
                        json.dumps(parameters),
                        endpoint_url,
                    ],
                    env=env,
                )

            return 0
        except OSError as ex:
            if ex.errno == errno.ENOENT:
                logger.debug(
                    'SessionManagerPlugin is not present', exc_info=True
                )
                # start-session api call returns response and starts the
                # session on ssm-agent and response is forwarded to
                # session-manager-plugin. If plugin is not present, terminate
                # is called so that service and ssm-agent terminates the
                # session to avoid zombie session active on ssm-agent for
                # default self terminate time
                client.terminate_session(SessionId=session_id)
                raise ValueError(''.join(ERROR_MESSAGE))



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import logging
import os
import sys

from botocore.utils import (
    generate_login_cache_key,
    get_login_token_cache_directory,
)

from awscli.customizations.commands import BasicCommand
from awscli.customizations.sso.logout import BaseCredentialSweeper
from awscli.customizations.utils import uni_print

LOG = logging.getLogger(__name__)


class LogoutCommand(BasicCommand):
    NAME = 'logout'
    DESCRIPTION = 'Clears cached login credentials for the specified profile.'
    ARG_TABLE = [
        {
            'name': 'all',
            'action': 'store_true',
            'help_text': (
                'Removes cached credentials for all profiles that use '
                '"aws login". Does not remove credentials from profiles '
                'configured to use other styles of credentials.'
            ),
        },
    ]

    def __init__(
        self,
        session,
        cache_dir=get_login_token_cache_directory(),
    ):
        super().__init__(session)
        self._cache_dir = cache_dir

    def _run_main(self, parsed_args, parsed_globals):
        if getattr(parsed_args, 'all', False):
            return self._logout_all()
        else:
            return self._logout_single_profile()

    def _logout_single_profile(self):
        profile = self._resolve_profile_name()
        session_id = self._get_login_session_id(profile)
        if session_id is None:
            msg = f"warning: no login session found for profile '{profile}'\n"
            uni_print(msg, sys.stderr)
            return 0

        cache_key = generate_login_cache_key(session_id)
        file_to_delete = os.path.join(self._cache_dir, f"{cache_key}.json")

        if os.path.exists(file_to_delete):
            os.remove(file_to_delete)
        uni_print(
            f"Removed cached login credentials for profile '{profile}'. "
            "Note, any local developer tools that have already loaded the "
            "access token may continue to use it until its expiration. "
            "Access tokens expire in 15 minutes.\n"
        )
        return 0

    def _logout_all(self):
        sweeper = LoginTokenSweeper()
        deleted_count = sweeper.delete_credentials(self._cache_dir)
        if deleted_count > 0:
            plural = "s" if deleted_count != 1 else ""
            uni_print(
                f"Removed {deleted_count} cached login credential{plural}. "
                "Note, any local developer tools that have already loaded "
                "access tokens may continue to use them until their "
                "expiration. Access tokens expire in 15 minutes.\n"
            )
        else:
            uni_print("No cached login session tokens found.\n")
        return 0

    def _resolve_profile_name(self):
        profile = self._session.profile
        return profile if profile is not None else 'default'

    def _get_login_session_id(self, profile):
        loaded_config = self._session.full_config
        profiles = loaded_config.get('profiles', {})
        profile_config = profiles.get(profile, {})
        return profile_config.get('login_session')


class LoginTokenSweeper(BaseCredentialSweeper):
    def delete_credentials(self, creds_dir):
        """Override to return count of deleted credentials."""
        if not os.path.isdir(creds_dir):
            return 0

        deleted_count = 0
        filenames = os.listdir(creds_dir)
        for filename in filenames:
            filepath = os.path.join(creds_dir, filename)
            contents = self._get_json_contents(filepath)
            if contents is None:
                continue
            if self._should_delete(contents):
                self._before_deletion(contents)
                os.remove(filepath)
                deleted_count += 1

        return deleted_count

    def _should_delete(self, contents):
        access_token = contents.get('accessToken')
        return (
            isinstance(access_token, dict) and 'accessKeyId' in access_token
        )



# Copyright 2025 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import json
import logging
import sys

import botocore

from awscli.customizations.commands import BasicCommand
from awscli.customizations.emrcontainers.base36 import Base36
from awscli.customizations.emrcontainers.constants import (
    SERVICE_ACCOUNT_NAMING,
    ServiceAccount,
)
from awscli.customizations.emrcontainers.eks import EKS
from awscli.customizations.emrcontainers.utils import get_region
from awscli.customizations.utils import get_policy_arn_suffix, uni_print

LOG = logging.getLogger(__name__)


class CreateRoleAssociationsCommand(BasicCommand):
    NAME = "create-role-associations"

    DESCRIPTION = BasicCommand.FROM_FILE(
        "emr-containers", "create-role-associations", "_description.rst"
    )

    ARG_TABLE = [
        {
            "name": "cluster-name",
            "help_text": (
                "Specify the name of the Amazon EKS cluster with "
                "which the IAM Role would be associated."
            ),
            "required": True,
        },
        {
            "name": "namespace",
            "help_text": (
                "Specify the job/application namespace from the Amazon EKS cluster "
                "with which the IAM Role would be associated."
            ),
            "required": True,
        },
        {
            "name": "role-name",
            "help_text": (
                "Specify the IAM Role name that you want to associate "
                "with Amazon EMR on EKS."
            ),
            "required": True,
        },
        {
            "name": "type",
            "help_text": (
                "Specify the Amazon EMR on EKS submission model and choose service accounts that you want to associate "
                "with Amazon EMR on EKS. The default is start_job_run. Supported types: start_job_run, "
                "interactive_endpoint, spark_operator, flink_operator, livy."
            ),
            "required": False,
            "choices": [
                "start_job_run",
                "interactive_endpoint",
                "spark_operator",
                "flink_operator",
                "livy",
            ],
        },
        {
            "name": "operator-namespace",
            "help_text": (
                "Specify the namespace that you want to associate the operator service account "
                "with the IAM role. Default to the job/application namespace specified. Note: "
                "If jobs are running in a different namespace than the operator installation namespace, "
                "this parameter needs to be set as the namespace that the operator is running on."
            ),
            "required": False,
        },
        {
            "name": "service-account-name",
            "help_text": (
                "Specify the service account name that you want to associate with the IAM role. "
                "By default, Amazon EMR on EKS service accounts will be used for association."
            ),
            "required": False,
        },
    ]

    def _run_main(self, parsed_args, parsed_globals):
        """Call to run the commands"""

        self._cluster_name = parsed_args.cluster_name
        self._namespace = parsed_args.namespace
        self._role_name = parsed_args.role_name
        self._type = parsed_args.type or "start_job_run"
        self._operator_namespace = parsed_args.operator_namespace
        self._service_account_name = parsed_args.service_account_name
        self._region = get_region(self._session, parsed_globals)

        result = self._create_role_associations(parsed_globals)
        if result:
            uni_print(json.dumps(result, indent=4))

        return 0

    def _create_role_associations(self, parsed_globals):
        """Method to create role associations if not done already"""
        eks_client = EKS(
            self._session.create_client(
                "eks",
                region_name=self._region,
                verify=parsed_globals.verify_ssl,
            )
        )
        account_id = eks_client.get_account_id(self._cluster_name)
        role_arn = f"arn:{get_policy_arn_suffix(self._region)}:iam::{account_id}:role/{self._role_name}"

        results = []
        # If service account provided, create association with provided service account and role
        if self._service_account_name:
            service_account_namespace_mapping = [
                (self._service_account_name, self._namespace)
            ]
        else:
            # By default, create associations with EMR on EKS service accounts
            base36 = Base36()

            base36_encoded_role_name = base36.encode(self._role_name)
            LOG.debug(f"Base36 encoded role name: {base36_encoded_role_name}")
            service_account_namespace_mapping = (
                self._get_emr_service_account_namespace_mapping(
                    account_id, base36_encoded_role_name
                )
            )

        for (
            service_account_name,
            namespace,
        ) in service_account_namespace_mapping:
            LOG.debug(
                f"Creating pod identity association for service account {service_account_name} and "
                + f"role {self._role_name} in namespace {namespace}"
            )
            try:
                result = eks_client.create_pod_identity_association(
                    self._cluster_name,
                    namespace,
                    role_arn,
                    service_account_name,
                )
                results.append(
                    result["association"]
                    if "association" in result
                    else result
                )
            except botocore.exceptions.ClientError as error:
                # Raise the error if EKS throws exceptions other than ResourceInUseException
                if error.response["Error"]["Code"] != "ResourceInUseException":
                    for result in results:
                        uni_print(
                            f"Rolling back association for service account {service_account_name} "
                            f"and role {self._role_name} in namespace {namespace} as an error was encountered\n",
                            out_file=sys.stderr,
                        )
                        eks_client.delete_pod_identity_association(
                            cluster_name=result["clusterName"],
                            association_id=result["associationId"],
                        )
                    raise Exception(
                        f"Failed to create pod identity association for service account {service_account_name}, "
                        f"role {self._role_name} in namespace {namespace}: {error.response['Error']['Message']}"
                    ) from error
                uni_print(
                    f"Skipping pod identity association creation because pod identity association already exists for "
                    f"service account {service_account_name} and role {self._role_name} in namespace {namespace}: "
                    f"{error.response['Error']['Message']}\n",
                    out_file=sys.stderr,
                )
        return results

    def _get_emr_service_account_namespace_mapping(
        self, account_id, base36_encoded_role_name
    ):
        return getattr(self, f"_get_{self._type}_mapping")(
            account_id, base36_encoded_role_name
        )

    def _get_start_job_run_mapping(self, account_id, base36_encoded_role_name):
        emr_spark_components = ["client", "driver", "executor"]
        return [
            (
                SERVICE_ACCOUNT_NAMING
                % {
                    "FRAMEWORK": "spark",
                    "COMPONENT": component,
                    "AWS_ACCOUNT_ID": account_id,
                    "BASE36_ENCODED_ROLE_NAME": base36_encoded_role_name,
                },
                self._namespace,
            )
            for component in emr_spark_components
        ]

    def _get_interactive_endpoint_mapping(
        self, account_id, base36_encoded_role_name
    ):
        emr_spark_endpoint_components = ["jeg", "jeg-kernel", "session"]
        return [
            (
                SERVICE_ACCOUNT_NAMING
                % {
                    "FRAMEWORK": "spark",
                    "COMPONENT": component,
                    "AWS_ACCOUNT_ID": account_id,
                    "BASE36_ENCODED_ROLE_NAME": base36_encoded_role_name,
                },
                self._namespace,
            )
            for component in emr_spark_endpoint_components
        ]

    def _get_spark_operator_mapping(
        self, account_id, base36_encoded_role_name
    ):
        emr_spark_operator_components = ["driver", "executor"]
        service_accounts = [
            (
                ServiceAccount.SPARK_OPERATOR_SERVICE_ACCOUNT.value,
                (
                    self._operator_namespace
                    if self._operator_namespace
                    else self._namespace
                ),
            )
        ]
        service_accounts.extend(
            [
                (
                    SERVICE_ACCOUNT_NAMING
                    % {
                        "FRAMEWORK": "spark",
                        "COMPONENT": component,
                        "AWS_ACCOUNT_ID": account_id,
                        "BASE36_ENCODED_ROLE_NAME": base36_encoded_role_name,
                    },
                    self._namespace,
                )
                for component in emr_spark_operator_components
            ]
        )
        return service_accounts

    def _get_flink_operator_mapping(
        self, account_id, base36_encoded_role_name
    ):
        emr_flink_operator_components = ["jobmanager", "taskmanager"]
        service_accounts = [
            (
                ServiceAccount.FLINK_OPERATOR_SERVICE_ACCOUNT.value,
                (
                    self._operator_namespace
                    if self._operator_namespace
                    else self._namespace
                ),
            )
        ]
        service_accounts.extend(
            [
                (
                    SERVICE_ACCOUNT_NAMING
                    % {
                        "FRAMEWORK": "flink",
                        "COMPONENT": component,
                        "AWS_ACCOUNT_ID": account_id,
                        "BASE36_ENCODED_ROLE_NAME": base36_encoded_role_name,
                    },
                    self._namespace,
                )
                for component in emr_flink_operator_components
            ]
        )
        return service_accounts

    def _get_livy_mapping(self, account_id, base36_encoded_role_name):
        return [
            (
                ServiceAccount.LIVY_SERVICE_ACCOUNT.value,
                (
                    self._operator_namespace
                    if self._operator_namespace
                    else self._namespace
                ),
            ),
            (ServiceAccount.LIVY_SPARK_SERVICE_ACCOUNT.value, self._namespace),
        ]



# language governing permissions and limitations under the License.
"""
Top Level Boolean Parameters
----------------------------

This customization will take a parameter that has
a structure of a single boolean element and allow the argument
to be specified without a value.

Instead of having to say::

    --ebs-optimized '{"Value": true}'
    --ebs-optimized '{"Value": false}'

You can instead say `--ebs-optimized/--no-ebs-optimized`.


"""

import logging
from functools import partial

from awscli import arguments
from awscli.argprocess import detect_shape_structure
from awscli.customizations.exceptions import ParamValidationError
from awscli.customizations.utils import validate_mutually_exclusive_handler

LOG = logging.getLogger(__name__)
# This sentinel object is used to distinguish when
# a parameter is not specified vs. specified with no value
# (a value of None).
_NOT_SPECIFIED = object()


def register_bool_params(event_handler):
    event_handler.register(
        'building-argument-table.ec2.*',
        partial(pull_up_bool, event_handler=event_handler),
    )


def _qualifies_for_simplification(arg_model):
    if detect_shape_structure(arg_model) == 'structure(scalar)':
        members = arg_model.members
        if (
            len(members) == 1
            and list(members.keys())[0] == 'Value'
            and list(members.values())[0].type_name == 'boolean'
        ):
            return True
    return False


def pull_up_bool(argument_table, event_handler, **kwargs):
    # List of tuples of (positive_bool, negative_bool)
    # This is used to validate that we don't specify
    # an --option and a --no-option.
    boolean_pairs = []
    event_handler.register(
        'operation-args-parsed.ec2.*',
        partial(validate_boolean_mutex_groups, boolean_pairs=boolean_pairs),
    )
    for value in list(argument_table.values()):
        if hasattr(value, 'argument_model'):
            arg_model = value.argument_model
            if _qualifies_for_simplification(arg_model):
                # Swap out the existing CLIArgument for two args:
                # one that supports --option and --option <some value>
                # and another arg of --no-option.
                new_arg = PositiveBooleanArgument(
                    value.name,
                    arg_model,
                    value._operation_model,
                    value._event_emitter,
                    group_name=value.name,
                    serialized_name=value._serialized_name,
                )
                argument_table[value.name] = new_arg
                negative_name = 'no-%s' % value.name
                negative_arg = NegativeBooleanParameter(
                    negative_name,
                    arg_model,
                    value._operation_model,
                    value._event_emitter,
                    action='store_true',
                    dest='no_%s' % new_arg.py_name,
                    group_name=value.name,
                    serialized_name=value._serialized_name,
                )
                argument_table[negative_name] = negative_arg
                # If we've pulled up a structure(scalar) arg
                # into a pair of top level boolean args, we need
                # to validate that a user only provides the argument
                # once.  They can't say --option/--no-option, nor
                # can they say --option --option Value=false.
                boolean_pairs.append((new_arg, negative_arg))


def validate_boolean_mutex_groups(boolean_pairs, parsed_args, **kwargs):
    # Validate we didn't pass in an --option and a --no-option.
    for positive, negative in boolean_pairs:
        if (
            getattr(parsed_args, positive.py_name) is not _NOT_SPECIFIED
            and getattr(parsed_args, negative.py_name) is not _NOT_SPECIFIED
        ):
            raise ParamValidationError(
                'Cannot specify both the "%s" option and '
                'the "%s" option.' % (positive.cli_name, negative.cli_name)
            )


class PositiveBooleanArgument(arguments.CLIArgument):
    def __init__(
        self,
        name,
        argument_model,
        operation_model,
        event_emitter,
        serialized_name,
        group_name,
    ):
        super(PositiveBooleanArgument, self).__init__(
            name,
            argument_model,
            operation_model,
            event_emitter,
            serialized_name=serialized_name,
        )
        self._group_name = group_name

    @property
    def group_name(self):
        return self._group_name

    def add_to_parser(self, parser):
        # We need to support three forms:
        # --option-name
        # --option-name Value=(true|false)
        parser.add_argument(
            self.cli_name,
            help=self.documentation,
            action='store',
            default=_NOT_SPECIFIED,
            nargs='?',
        )

    def add_to_params(self, parameters, value):
        if value is _NOT_SPECIFIED:
            return
        elif value is None:
            # Then this means that the user explicitly
            # specified this arg with no value,
            # e.g. --boolean-parameter
            # which means we should add a true value
            # to the parameters dict.
            parameters[self._serialized_name] = {'Value': True}
        else:
            # Otherwise the arg was specified with a value.
            parameters[self._serialized_name] = self._unpack_argument(value)


class NegativeBooleanParameter(arguments.BooleanArgument):
    def __init__(
        self,
        name,
        argument_model,
        operation_model,
        event_emitter,
        serialized_name,
        action='store_true',
        dest=None,
        group_name=None,
    ):
        super(NegativeBooleanParameter, self).__init__(
            name,
            argument_model,
            operation_model,
            event_emitter,
            default=_NOT_SPECIFIED,
            serialized_name=serialized_name,
        )
        self._group_name = group_name

    def add_to_params(self, parameters, value):
        if value is not _NOT_SPECIFIED and value:
            parameters[self._serialized_name] = {'Value': False}



# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import sys

from awscli.customizations.commands import BasicCommand

from . import NOT_SET, ConfigValue


class ConfigureListCommand(BasicCommand):
    NAME = 'list'
    DESCRIPTION = (
        'Lists the profile, access key, secret key, and region configuration '
        'information used for the specified profile. For each configuration '
        'item, it shows the value, where the configuration value '
        'was retrieved, and the configuration variable name.\n'
        '\n'
        'For example, '
        'if you provide the AWS region in an environment variable, this '
        'command shows you the name of the region you\'ve configured, '
        'that this value came from an environment '
        'variable, and the name of the environment '
        'variable.\n'
        '\n'
        'For temporary credential methods such as roles and IAM Identity '
        'Center, this command displays the temporarily cached access key and '
        'secret access key is displayed.\n'
    )
    SYNOPSIS = 'aws configure list [--profile profile-name]'
    EXAMPLES = (
        'To show your current configuration values::\n'
        '\n'
        '  $ aws configure list\n'
        '  NAME       : VALUE                    : TYPE             : LOCATION\n'
        '  profile    : <not set>                : None             : None\n'
        '  access_key : ****************ABCD     : config_file      : ~/.aws/config\n'
        '  secret_key : ****************ABCD     : config_file      : ~/.aws/config\n'
        '  region     : us-west-2                : env              : AWS_DEFAULT_REGION\n'
        '\n'
    )

    def __init__(self, session, stream=None):
        super().__init__(session)
        if stream is None:
            stream = sys.stdout
        self._stream = stream

    def _run_main(self, args, parsed_globals):
        self._display_config_value(
            ConfigValue('VALUE', 'TYPE', 'LOCATION'), 'NAME'
        )

        if parsed_globals and parsed_globals.profile is not None:
            profile = ConfigValue(self._session.profile, 'manual', '--profile')
        else:
            profile = self._lookup_config('profile')
        self._display_config_value(profile, 'profile')

        access_key, secret_key = self._lookup_credentials()
        self._display_config_value(access_key, 'access_key')
        self._display_config_value(secret_key, 'secret_key')

        region = self._lookup_region()
        self._display_config_value(region, 'region')
        return 0

    def _display_config_value(self, config_value, config_name):
        self._stream.write(
            f'{config_name:<10} : '
            f'{config_value.value:<24} : '
            f'{str(config_value.config_type):<16} : '
            f'{str(config_value.config_variable)}\n'
        )

    def _lookup_credentials(self):
        # First try it with _lookup_config.  It's possible
        # that we don't find credentials this way (for example,
        # if we're using an IAM role).
        access_key = self._lookup_config('access_key')
        if access_key.value is not NOT_SET:
            secret_key = self._lookup_config('secret_key')
            access_key.mask_value()
            secret_key.mask_value()
            return access_key, secret_key
        else:
            # Otherwise we can try to use get_credentials().
            # This includes a few more lookup locations
            # (IAM roles, some of the legacy configs, etc.)
            credentials = self._session.get_credentials()
            if credentials is None:
                no_config = ConfigValue(NOT_SET, None, None)
                return no_config, no_config
            else:
                # For the ConfigValue, we don't track down the
                # config_variable because that info is not
                # visible from botocore.credentials.  I think
                # the credentials.method is sufficient to show
                # where the credentials are coming from.
                access_key = ConfigValue(
                    credentials.access_key, credentials.method, ''
                )
                secret_key = ConfigValue(
                    credentials.secret_key, credentials.method, ''
                )
                access_key.mask_value()
                secret_key.mask_value()
                return access_key, secret_key

    def _lookup_region(self):
        region = self._lookup_in_env_and_config('region')
        if region:
            return region
        region = self._session.get_config_variable('region')
        if region:
            # This makes the assumption that the only other possible place to
            # retrieve region in the CLI is via IMDS. If we add more region
            # providers in the future this will have to be updated.
            #
            # TODO: Currently credentials use the value iam-role, but the
            # region will return imds as region is not included in iam-role.
            # Ideally both imds and credentials should return the same source
            # value back if they both came from IMDS.
            return ConfigValue(region, 'imds', '')
        else:
            return ConfigValue(NOT_SET, None, None)

    def _lookup_in_env_and_config(self, name):
        # First try to look up the variable in the env.
        value = self._session.get_config_variable(name, methods=('env',))
        if value is not None:
            return ConfigValue(
                value, 'env', self._session.session_var_map[name][1]
            )
        # Then try to look up the variable in the config file.
        value = self._session.get_config_variable(name, methods=('config',))
        if value is not None:
            return ConfigValue(
                value,
                'config-file',
                self._session.get_config_variable('config_file'),
            )

    def _lookup_config(self, name):
        val = self._lookup_in_env_and_config(name)
        if val:
            return val
        else:
            return ConfigValue(NOT_SET, None, None)
